package Kpathsea;

use 5.008;
use strict;
use warnings;

our $VERSION = '0.01';
our $KPATHSEA_VERSION = '3.4.5';

use FindBin;
use Data::Dumper;

sub new {
  my ($class,$rootdir,$tmpdir) = @_;
  my $this = {};
  bless $this,$class;
  $this->{kpathsea_debug} = 0;
  $this->fixup_opsys();
  $this->{KPRE} = $rootdir || "/usr/share";
  $this->{KTMP} = $tmpdir  || "/var/tmp/texfonts";
  $this->init_c_vars();
  $this->{link_table} = {};
  $this->{not_first_search}=0;
  return $this;
}

sub IS_DIR_SEP { 
  my ($this,$a) = @_; 
  return 1 if $a =~ /[$this->{DIR_SEP}$this->{DIR_SEPA}]/; 
  return 0;
}

sub IS_ENV_SEP { 
  my ($this,$a) = @_; 
  return 1 if $a eq $this->{ENV_SEP}; 
  return 0; 
}

sub IS_DEVICE_SEP { 
  my ($this,$a) = @_;
  return 1 if $a eq ':'  && $^O =~ /(msdos|win32|os2|amiga)/;
  return 0;
}

sub NAME_BEGINS_WITH_DEVICE { 
  my ($this,$a) = @_;
  return 1 if $^O =~ /(msdos|win32|os2)/ && $a =~ /^.:/;
  return 0;
}

sub IS_UNC_NAME { 
  my ($this,$a) = @_;
  return 1 if ($^O =~ /win32/ && length($a)>=3 && $a =~ /^([\/\\])\1[a-zA-Z0-9]/);
  return 0; 
}


sub fixup_opsys {
  my ($this) = @_;
  $this->{MONOCASE_FILENAMES} =0;
  $this->{NAME_MAX} = 255; # TODO!! 
  $this->{DIR_SEPA} = '';
  if ($^O =~ /vms/i) {
    $this->{DIR_SEP} = ':';
    $this->{ENV_SEP} = ',';
  } elsif ($^O =~ /(msdos|win32|os2)/i) { # dosish
    $this->{MONOCASE_FILENAMES} = 1;
    $this->{DIR_SEP} = '/';
    $this->{DIR_SEPA} = '\\';
    $this->{ENV_SEP} = ';';
  } elsif ($^O =~ /amiga/i) {
    $this->{DIR_SEP} = '/';
    $this->{DIR_SEPA} = ':';
    $this->{ENV_SEP} = ';';
  } elsif ($^O =~ /cms/i) {
    $this->{DIR_SEP} = ' ';
    $this->{ENV_SEP} = ' ';
  } else {
    $this->{DIR_SEP} = '/';
    $this->{ENV_SEP} = ':';
  }
}

sub perror {
  my ($this,$text) = @_;
  print STDERR "Kpathsea error: $! from " . $text . "\n";
}

#readable.h: readable_file()
#str-list.h: C lists of strings
#str-llist.h: C linked lists of strings
#systypes.h: internal __TYPES__ def

#tex-file.h: 

# the file_format_type enum typedef

sub init_c_vars {
  my $this = shift;
  $this->{program_invocation_name} = '';
  $this->{program_invocation_short_name} = '';
  $this->{program_name} = '';

  $this->{fallback_font} = '';
  $this->{fallback_resolutions_string} = '';
  $this->{fallback_resolutions} = [];
  
  my $file_format_type = 0;
  foreach my $word (qw(gf pk any_glyph tfm afm base bib bst cnf db fmt fontmap
		       mem mf mfpool mft mp mppool mpsupport
		       ocp ofm opl otp ovf ovp pict
		       tex texdoc texpool texsource tex_ps_header troff_font
		       type1 vf dvips_config ist truetype type42 web2c
		       program_text program_binary miscfonts web cweb
		      )) {
    $this->{$word . "_format"} =  $file_format_type++;
  }
  $this->{last_format} = $file_format_type;
  #
  my $src_type = 0;
  foreach my $word (qw(implicit compile texmf_cnf client_cnf env x cmdline)) {
    $this->{"src_" . $word} = $src_type++;
  }
  #
  $this->{format_info} = [];
  #
  my $glyph_source_type=0;
  foreach my $word (qw(normal alias maketex fallback)) {
    $this->{"glyph_source_" . $word} = $glyph_source_type++;
  }
  #
  $this->{CNF_NAME} = 'texmf.cnf';
  $this->{kpathsea_version_string} = "kpathsea version $KPATHSEA_VERSION, perl version $VERSION";
  $this->{bug_address} = "Email bug reports to taco\@elvenkind.com.\n";
  #
  $this->{DB_NAME} = 'ls-R';
  $this->{ALIAS_NAME} = 'aliases';
  $this->{MAP_NAME} = 'texfonts.map';
  $this->{UNIX_BITMAP_SPEC} = "\$KPATHSEA_NAME.\$KPATHSEA_DPI\$KPATHSEA_FORMAT";
  $this->{DPI_BITMAP_SPEC}  =  "dpi\$KPATHSEA_DPI/\$KPATHSEA_NAME.\$KPATHSEA_FORMAT";

  $this->{GLYPH_ENVS} = ["GLYPHFONTS", "TEXFONTS"];
  $this->{GF_ENVS} = ["GFFONTS", @{$this->{GLYPH_ENVS}}];
  $this->{PK_ENVS} = ["PKFONTS", "TEXPKS", @{$this->{GLYPH_ENVS}}];
  $this->{TFM_ENVS} = ["TFMFONTS", "TEXFONTS"];
  $this->{AFM_ENVS} = ["AFMFONTS"];
  $this->{BASE_ENVS} = ["MFBASES", "TEXMFINI"];
  $this->{BIB_ENVS} = ["BIBINPUTS", "TEXBIB"];
  $this->{BST_ENVS} = ["BSTINPUTS"];
  $this->{CNF_ENVS} = ["TEXMFCNF"];
  $this->{DB_ENVS} = ["TEXMFDBS"];
  $this->{FMT_ENVS} = ["TEXFORMATS", "TEXMFINI"];
  $this->{FONTMAP_ENVS} = ["TEXFONTMAPS"];
  $this->{MEM_ENVS} = ["MPMEMS", "TEXMFINI"];
  $this->{MF_ENVS} = ["MFINPUTS"];
  $this->{MFPOOL_ENVS} = ["MFPOOL", "TEXMFINI"];
  $this->{MFT_ENVS} = ["MFTINPUTS"];
  $this->{MP_ENVS} = ["MPINPUTS"];
  $this->{MPPOOL_ENVS} = ["MPPOOL", "TEXMFINI"];
  $this->{MPSUPPORT_ENVS} = ["MPSUPPORT"];
  $this->{OCP_ENVS} = ["OCPINPUTS"];
  $this->{OFM_ENVS} = ["OFMFONTS", "TEXFONTS"];
  $this->{OPL_ENVS} = ["OPLFONTS", "TEXFONTS"];
  $this->{OTP_ENVS} = ["OTPINPUTS"];
  $this->{OVF_ENVS} = ["OVFFONTS", "TEXFONTS"];
  $this->{OVP_ENVS} = ["OVPFONTS", "TEXFONTS"];
  $this->{TEX_ENVS} = ["TEXINPUTS"];
  $this->{PICT_ENVS} = ["TEXPICTS", @{$this->{TEX_ENVS}}];
  $this->{TEXDOC_ENVS} = ["TEXDOCS"];
  $this->{TEXPOOL_ENVS} = ["TEXPOOL", "TEXMFINI"];
  $this->{TEXSOURCE_ENVS} = ["TEXSOURCES"];
  $this->{TEX_PS_HEADER_ENVS} = ["TEXPSHEADERS", "PSHEADERS"];
  $this->{TROFF_FONT_ENVS} = ["TRFONTS"];
  $this->{TYPE1_ENVS} = ["T1FONTS", "T1INPUTS", @{$this->{TEX_PS_HEADER_ENVS}}];
  $this->{VF_ENVS} = ["VFFONTS", "TEXFONTS"];
  $this->{DVIPS_CONFIG_ENVS} = ["TEXCONFIG"];
  $this->{IST_ENVS} = ["TEXINDEXSTYLE", "INDEXSTYLE"];
  $this->{TYPE_ENVS} = ["TTFONTS"];
  $this->{TYPE42_ENVS} = ["T42FONTS"];
  $this->{WEB2C_ENVS} = ["WEB2C"];
  $this->{MISCFONTS_ENVS} = ["MISCFONTS"];
  $this->{WEB_ENVS} = ["WEBINPUTS"];
  $this->{CWEB_ENVS} = ["CWEBINPUTS"];
  $this->{DEFAULT_FONT_SIZES} = "600";
  $this->{MKTEXPK_ARGS}  = [ "--mfmode","\$MAKETEX_MODE","--bdpi","\$MAKETEX_BASE_DPI",
		       "--mag","\$MAKETEX_MAG", "--dpi","\$KPATHSEA_DPI" ];
  $this->xputenv("KPRE",$this->{KPRE});
  $this->xputenv("KTMP",$this->{KTMP});
  $this->{DEFAULTS}->{TEXMFMAIN} = "\$KPRE/texmf";
  $this->{DEFAULTS}->{TEXMFLOCAL} = "\$KPRE/texmf-local";
  $this->{DEFAULTS}->{HOMETEXMF} = "\$HOME/texmf";
  $this->{DEFAULTS}->{TEXMF} = "{\$HOMETEXMF,!!\$KPRE/texmf-local,!!\$KPRE/texmf}";
  $this->{DEFAULTS}->{SYSTEXMF} = "\$KPRE/texmf";
  $this->{DEFAULTS}->{VARTEXFONTS} = "\$KPTMP";
  $this->{DEFAULTS}->{TEXMFDBS} = "\$KPRE/texmf:\$KPTMP";
  $this->{DEFAULTS}->{WEB2C} = "\$KPRE/texmf/web2c";
  $this->{DEFAULTS}->{TEXINPUTS} = ".:\$KPRE/texmf/tex/{generic,}//";
  $this->{DEFAULTS}->{MFINPUTS} = ".:\$KPRE/texmf/metafont//:{\$KPRE/texmf/fonts,\$KPTMP}/source//";
  $this->{DEFAULTS}->{MPINPUTS} = ".:\$KPRE/texmf/metapost//";
  $this->{DEFAULTS}->{TEXFORMATS} = ".:\$KPRE/texmf/web2c";
  $this->{DEFAULTS}->{MFBASES} = ".:\$KPRE/texmf/web2c";
  $this->{DEFAULTS}->{MPMEMS} = ".:\$KPRE/texmf/web2c";
  $this->{DEFAULTS}->{TEXPOOL} = ".:\$KPRE/texmf/web2c";
  $this->{DEFAULTS}->{MFPOOL} = ".:\$KPRE/texmf/web2c";
  $this->{DEFAULTS}->{MPPOOL} = ".:\$KPRE/texmf/web2c";
  $this->{DEFAULTS}->{VFFONTS} = ".:\$KPRE/texmf/fonts/vf//";
  $this->{DEFAULTS}->{TFMFONTS} = ".:{\$KPRE/texmf/fonts,\$KPTMP}/tfm//";
  $this->{DEFAULTS}->{PKFONTS} = ".:{\$KPRE/texmf/fonts,\$KPTMP}/pk/{\$MAKETEX_MODE,modeless}//";
  $this->{DEFAULTS}->{GFFONTS} = ".:\$KPRE/texmf/fonts/gf/\$MAKETEX_MODE//";
  $this->{DEFAULTS}->{GLYPHFONTS} = ".:\$KPRE/texmf/fonts";
  $this->{DEFAULTS}->{MISCFONTS} = ".:\$KPRE/texmf/fonts/misc//";
  $this->{DEFAULTS}->{TEXFONTMAPS} = ".:\$KPRE/texmf/fontname";
  $this->{DEFAULTS}->{BIBINPUTS} = ".:\$KPRE/texmf/bibtex/{bib,}//";
  $this->{DEFAULTS}->{BSTINPUTS} = ".:\$KPRE/texmf/bibtex/{bst,}//";
  $this->{DEFAULTS}->{MFTINPUTS} = ".:\$KPRE/texmf/mft//";
  $this->{DEFAULTS}->{TEXPSHEADERS} = ".:\$KPRE/texmf/{dvips,pdftex,tex,fonts/type1}//";
  $this->{DEFAULTS}->{T1FONTS} = ".:\$KPRE/texmf/fonts/type1//";
  $this->{DEFAULTS}->{AFMFONTS} = ".:\$KPRE/texmf/fonts/afm//";
  $this->{DEFAULTS}->{TTFONTS} = ".:\$KPRE/texmf/fonts/truetype//";
  $this->{DEFAULTS}->{T42FONTS} =  ".:\$KPRE/texmf/fonts/type42//";
  $this->{DEFAULTS}->{TEXCONFIG} = ".:\$KPRE/texmf/dvips//";
  $this->{DEFAULTS}->{INDEXSTYLE} = ".:\$KPRE/texmf/makeindex//";
  $this->{DEFAULTS}->{TRFONTS} = "/usr/lib/font/devpost";
  $this->{DEFAULTS}->{MPSUPPORT} = ".:\$KPRE/texmf/metapost/support";
  $this->{DEFAULTS}->{MIMELIBDIR} = "\$KPRE/etc";
  $this->{DEFAULTS}->{MAILCAPLIBDIR} = "\$KPRE/etc";
  $this->{DEFAULTS}->{TEXDOCS} = ".:\$KPRE/texmf/doc//";
  $this->{DEFAULTS}->{TEXSOURCES} = ".:\$KPRE/texmf/source//";
  $this->{DEFAULTS}->{WEBINPUTS} = ".:\$KPRE/texmf/web//";
  $this->{DEFAULTS}->{CWEBINPUTS} = ".:\$KPRE/texmf/cweb//";
  $this->{DEFAULTS}->{OFMFONTS} = ".:{\$KPRE/texmf/fonts,\$KPTMP}/{ofm,tfm}//:\$TFMFONTS";
  $this->{DEFAULTS}->{OPLFONTS} = ".:{\$KPRE/texmf/fonts,\$KPTMP}/opl//";
  $this->{DEFAULTS}->{OVFFONTS} = ".:{\$KPRE/texmf/fonts,\$KPTMP}/ovf//";
  $this->{DEFAULTS}->{OVPFONTS} = ".:{\$KPRE/texmf/fonts,\$KPTMP}/ovp//";
  $this->{DEFAULTS}->{OTPINPUTS} = ".:\$KPRE/texmf/omega/otp//";
  $this->{DEFAULTS}->{OCPINPUTS} = ".:\$KPRE/texmf/omega/ocp//";
  $this->{DEFAULTS}->{T4HTINPUTS} = ".:\$KPRE/texmf/tex4ht//";
  $this->{DEFAULTS}->{XDVIINPUTS} = ".:\$KPRE/texmf/{xdvi,dvips//}";
  $this->{DEFAULTS}->{KPSE_DOT} = ".";
  $this->{DEFAULTS}->{TEXMFCNF} = "{\$SELFAUTOLOC,\$SELFAUTODIR,\$SELFAUTOPARENT}{,{/share,}/texmf{-local,}/web2c}:\$TETEXDIR:\$KPRE/texmf/web2c:\$KPRE/texmf/web2c";
  $this->{DEFAULTS}->{MISSFONT_LOG} = "missfont.log";
  $this->{DEFAULTS}->{TEX_HUSH} = "none";
  $this->{DEFAULTS}->{MPXCOMMAND} = "makempx";
}

# format_info_type = {
#   type => undef,
#   path => undef,
#   raw_path => undef,
#   path_source => undef,
#   override_path => undef,
#   client_path => undef,
#   cnf_path => undef,
#   default_path => undef,
#   suffix => undef,
#   alt_suffix => undef,
#   suffix_search_only => undef,
#   program => undef,
#   argc => undef,
#   argv => undef,
#   program_enabled_p => undef,
#   program_enable_level => undef,
#   binmode => undef,
# }

# glyph_file_type = {
#   name => undef,
#   dpi => undef,
#   format => undef,
#   source => undef,
#}

sub find_mf  { my ($this,$n) = @_; return $this->find_file ($n,$this->{mf_format},  1);}
sub find_mft { my ($this,$n) = @_; return $this->find_file ($n,$this->{mft_format}, 1);}
sub find_pict{ my ($this,$n) = @_; return $this->find_file ($n,$this->{pict_format},1);}
sub find_tex { my ($this,$n) = @_; return $this->find_file ($n,$this->{tex_format}, 1);}
sub find_tfm { my ($this,$n) = @_; return $this->find_file ($n,$this->{tfm_format}, 1);}
sub find_ofm { my ($this,$n) = @_; return $this->find_file ($n,$this->{ofm_format}, 1);}
sub find_vf  { my ($this,$n) = @_; return $this->find_file ($n,$this->{vf_format},  0);}
sub find_ovf { my ($this,$n) = @_; return $this->find_file ($n,$this->{ovf_format}, 0);}

sub find_pk {
  my ($this,$font_name, $dpi) = @_;
  return $this->find_glyph ($font_name, $dpi, $this->{pk_format});
}

sub find_gf {
  my ($this,$font_name, $dpi) = @_;
  return $this->find_glyph ($font_name, $dpi, $this->{gf_format});
}

sub absolute_p {
  my ($this,$filename, $relative_ok) = @_;
  return 1 if $filename =~ /[\]\>:]/ && $^O =~ /vms/i;
  return 0 if  $^O =~ /vms/i;
  my $absolute = $this->IS_DIR_SEP(substr($filename,0,1));
  unless ($absolute) {
	if ($^O =~ /(msdos|win32|os2)/i) {
	  $absolute = 1 if scalar($filename)>1 && $this->IS_DEVICE_SEP(substr($filename,1,1));
	} elsif ($^O =~ /amiga/i) {
	  $absolute  = 1 if $filename =~ /:/;
	}
  }
  my $explicit_relative = $relative_ok;
  if ($explicit_relative) {
	$explicit_relative = 0;
	my $chara = substr($filename,0,1);
	if ($^O =~ /amiga/i) {
	  $explicit_relative = 1 if $this->IS_DIR_SEP ($chara);
	}
	my $charb = substr($filename,1,1);
	my $charc = substr($filename,2,1);
	$explicit_relative = 1 
	  if $chara eq '.' && ($this->IS_DIR_SEP($charb) || 
			       ($charb eq '.' && $this->IS_DIR_SEP($charc))
			      );
      }
  return $absolute || $explicit_relative;
}


#line.c

sub read_line {
  my ($this,$f) = @_;
  my $line = <$f>;
  return $line if defined $line;
  return undef;
}

sub do_line {
  my ($this,$line) = @_;
  $line =~ s/^[ \t]+//;
  $line =~ s/[\#%].*$//;
  return unless $line;
  return if $line =~ /^[\#%]/;
  $line =~ /^([^=. \t]+)[ \t]*\.?[ \t]*([^=. ]+)?[ \t]*=?[ \t]*([^ \t]*)/;
  my $var  = $1;
  my $prog = $2;
  my $val  = $3;
  if ($this->IS_ENV_SEP(':') && !defined $this->{ALLOW_SEMICOLON_IN_FILENAMES}) {
    $val =~ tr /;/:/;
  }
  if (defined $prog && $prog) {
    $this->{cnf_hash}->{"$var.$prog"} = $val unless defined $this->{cnf_hash}->{"$var.$prog"};
  } else {
    $this->{cnf_hash}->{$var} = $val unless defined $this->{cnf_hash}->{$var} ;
  }
}

sub read_all_cnf {
  my ($this) = @_;
  my $cnf_path = $this->init_format ($this->{cnf_format});
  foreach my $cnf_filename (@{$this->all_path_search ($cnf_path, $this->{CNF_NAME})}) {
    my $cnf_file = $this->xfopen($cnf_filename,"<");
    while (my $line = $this->read_line($cnf_file)) {
      chop $line;
      $line =~ s/[ \t]*$//;
      while ($line =~ /\\$/) {
	my $xline = $this->read_line($cnf_file);
	if (defined $xline) {
	  chop $line; # kill the "\" char
	  chop $xline;
	  $xline =~ s/[ \t]*$//;
	  $line .= $xline;
	} else {
	  # error!!
	  chop $line; # kill the "\" char
	}
      }
      $this->do_line($line);
    }
    $this->xfclose($cnf_file);
  }
}

sub cnf_get {
  my ($this,$name) = @_;
  return undef if $this->{doing_cnf_init};
  if (!defined $this->{cnf_hash}) {
    $this->{doing_cnf_init} = 1;
    $this->read_all_cnf();
    $this->{doing_cnf_init} = 0;
    $this->init_db();
  }
  my $var;
  $var = $this->hash_lookup('cnf_hash',"$name." . $this->{program_name});
  unless (defined $var && $var) {
    $var = $this->hash_lookup('cnf_hash',$name);
  }
  return $var->[0];
}


# bits:
my $DEBUG_STAT   =0;  #           /* stat calls */
my $DEBUG_HASH   =1;  #           /* hash lookups */
my $DEBUG_FOPEN  =2;  #           /* fopen/fclose calls */
my $DEBUG_PATHS  =3;  #           /* search path initializations */
my $DEBUG_EXPAND =4;  #           /* path element expansion */
my $DEBUG_SEARCH =5;  #           /* searches */
my $DEBUG_VARS   =6;  #           /* variable values */

my $LAST_DEBUG   = $DEBUG_VARS;

sub DEBUG_SET { 
  my ($this,$bit) = @_; 
  $this->{kpathsea_debug} |= (1<<$bit); 
}

sub DEBUG_P { 
  my ($this,$bit) = @_; 
  return $this->{kpathsea_debug} & (1<<$bit); 
}

sub DEBUG {
  my $warning = shift;
  print STDERR "kdebug:$warning";
}

sub DEB {
  my $warning = shift;
  print STDERR "$warning";
}

sub DEBUGF {
  my ($warning,@vals) = @_;
  my $s = sprintf("kdebug:$warning",@vals);
  print STDERR $s;
}

sub dir_links {
  my ($this,$fn) = @_;
  my $nlinks = $this->{link_table}->{$fn};
  if (!defined $nlinks) {
    $this->{link_table}->{$fn} = (stat($fn))[3];
    $this->{link_table}->{$fn} = -1 unless -d $fn;
    $nlinks = $this->{link_table}->{$fn};
    if ($this->DEBUG_P ($DEBUG_STAT)) {
      DEBUG ("dir_links($fn) => $nlinks\n");
    }
  }
  return $nlinks;
}

#find-suffix.c

sub find_suffix {
  my ($this,$name) = @_;
  my $chartest = "$this->{DIR_SEP}$this->{DIR_SEPA}.";
  $name =~ /\.([^$chartest])$/;
  return $1;
}

sub extend_filename {
  my ($this,$name,$default_suffix) = @_;
  my $chartest = "$this->{DIR_SEP}$this->{DIR_SEPA}.";
  return $name if $name =~ /\.([^$chartest])$/;
  return "$name.$default_suffix";
}

sub remove_suffix {
  my ($this,$name) = @_;
  my $chartest = "$this->{DIR_SEP}$this->{DIR_SEPA}.";
  $name =~ s/^\.([^$chartest])$//;
  return $name;
}

sub make_suffix {
  my ($this,$name,$suffix) = @_;
  my $chartest = "$this->{DIR_SEP}$this->{DIR_SEPA}";
  $name =~ s/\.([^$chartest])$/.$suffix/;
  return $name;
}

sub same_file_p {
 my ($filea,$fileb) = @_;
 my @sta = stat($filea) || (-1,-1);
 my @stb = stat($fileb)  || (-2,-2);
 return ($sta[0] == $stb[0] && $sta[1] == $stb[1] ? 1 : 0);
}

sub hash_insert_normalized {
  my ($this,$table,$key,$value) = @_;
  push @{$this->{$table}->{$key}}, $value;
}

sub hash_print {
  my ($this,$table,$summary_only) = @_;
  my $b;
  my $total_elements = 0;
  my $total_buckets = 0;
  my $tablesize = 0;
  foreach my $key (keys %{$this->{$table}}) {
    $total_buckets++;
    $tablesize++;
    my $contents = @{$this->{$table}->{$key}};
    $total_elements += scalar($contents);
  }
  my $chain = ($total_buckets ? $total_elements /  $total_buckets : 0.0);
  $chain = int ($chain*10); $chain = $chain/10;
  print STDERR "$tablesize buckets, $total_buckets nonempty (100%); $total_elements entries, average chain $chain.\n";
}

  

sub hash_insert {
  my ($this,$table,$key,$value) = @_;
  if ($this->{MONOCASE_FILENAMES}) {
    push @{$this->{$table}->{lc($key)}}, $value;
  } else {
    push @{$this->{$table}->{$key}}, $value;
  }
}

sub hash_lookup {
  my ($this,$table,$key) = @_;
  my $ret = undef;
  if ($this->{MONOCASE_FILENAMES}) {
    $key = lc($key);
  }
  $ret= $this->{$table}->{$key} if defined $this->{$table}->{$key};
  $ret = [$ret] if defined $ret && ref($ret) ne 'ARRAY';
  if ($this->DEBUG_P ($DEBUG_HASH)) {
    DEBUG ("hash_lookup($key) =>");
    if (!defined $ret) {
      DEB(" (nil)\n");
    } else {
      foreach my $val (@$ret) {
	DEB (" $val");
      }
      DEB("\n");
    }
  }
  return $ret;
}

sub readable_file {
  my ($this,$name) = @_;
  $name = $this->normalize_path($name);
  if (-r $name && !-d $name) {
	return $name;
  } elsif ($! eq 'ENAMETOOLONG') {
	$name = $this->truncate_filename($name);
	if (-r $name && !-d $name) {
	  return $name;
	} else {
	  return undef;
	}
  } else {
    if ($! eq 'EACCES') {
      if (!$this->tex_hush ("readable")) {
	$this->perror ($name);
      }
    }
    return undef;
  }
}


#truncate.c


sub truncate_filename {
  my ($this,$name) = @_;
  my @chars = split(//,$name);
  my $c_len = 0;
  my $ret = '';
  foreach my $char (@chars) {
    if ($this->IS_DIR_SEP($char) || $this->IS_DEVICE_SEP($char)) {
      $c_len = 0;
    } elsif ($c_len> $this->{NAME_MAX}) {
      next;
    }
    $ret .= $char;
    $c_len++;
  }
  return $ret;
}

sub xopendir {
  my ($this,$dirhandle,$dir) = @_;
  if (opendir($dirhandle,$dir)) {
    return $dirhandle;
  } else {
    $this->perror("Opendir $dir failed\n");
    exit 1;
  }
}

sub xclosedir {
  my ($this,$dirhandle) = @_;
  if (closedir($dirhandle)) {
  } else {
    $this->perror("Closedir failed\n");
    exit 1;
  }
}

#xfopen.c


sub xfopen {
  my ($this,$file,$mode) = @_;
  my $filehandle;
  my $retval =  open($filehandle,"$mode$file");
  if ($this->DEBUG_P ($DEBUG_FOPEN)) {
    DEBUG("fopen($file,$mode) => $filehandle ($retval)\n");
  }
  return $filehandle if $retval;
  $this->perror("Fileopen $mode$file failed\n");
  exit 1;
}

sub xfclose {
  my ($this,$filehandle,$filename) = @_;
  my $retval = close($filehandle);
  if ($this->DEBUG_P ($DEBUG_FOPEN)) {
    DEBUG("fclose($filehandle) => $retval\n");
  }
  return 1 if $retval;
  $this->perror("Closefile $filename failed\n");
  exit 1;
}


#tex-hush.c

sub tex_hush {
  my ($this,$what) = @_;
  my $hush = $this->var_value('TEX_HUSH');
  return 1 if $hush eq 'all';
  return 0 if $hush eq 'none';
  foreach my $el ($this->path_elements($hush)) {
    return 1 if $el eq $what;
  }
  return 0;
}

#xputenv.c

sub xputenv {
  my ($this,$var,$value) = @_;
  $ENV{$var} = "$value";
}

#magstep.c

sub magstep {
  my ($this,$n,$bdpi) = @_;
  my ($t,$step,$neg);
  if ($n < 0)    { $neg = 1; $n = -$n; }
  if ($n & 1)    { $n &= ~1; $t = 1.095445115;  }
  else           { $t = 1.0; }
  while ($n > 8) { $n -= 8; $t = $t * 2.0736; }
  while ($n > 0) { $n -= 2; $t = $t * 1.2; }
  $step = int(0.5 + ($neg ? $bdpi / $t : $bdpi * $t));
  return $step;
}

# $m_ret is an adres!
sub magstep_fix {
  my ($this,$dpi,$bdpi,$m_ret) = @_;
  my $m;
  my $mdpi = -1;
  my $real_dpi = 0;
  my $sign = $dpi < $bdpi ? -1 : 1;
  for ($m = 0; $m < 40; $m++) { # /* don't go forever */
	$mdpi = $this->magstep ($m * $sign, $bdpi);
	if (abs ($mdpi - int ($dpi)) <= 1) { # /* if this magstep matches, quit */
	  $real_dpi = $mdpi; 
	} elsif (($mdpi - int ($dpi)) * $sign > 0) { # /* if gone too far, quit */
	  $real_dpi = $dpi; 
	}
	last if $real_dpi;
  }
  #/* If requested, return the encoded magstep (the loop went one too far).  */
  #/* More unnecessary casts. */
  if ($$m_ret) {
    $$m_ret = ($real_dpi == $mdpi ? ($m - 1) * $sign : 0);
  }
  # /* Always return the true dpi found.  */
  return $real_dpi ? $real_dpi : $dpi;
}

#path-elt.c


sub path_elements {
  my ($this,$path) = @_;
  return undef unless defined $path;
  my @things=();
  my $current = '';
  my $bracelevel = 0;
  my @text = split(/(\$|\{|\}|$this->{ENV_SEP})/,$path);
  while (@text) {
    my $item = shift @text;
    if ($item eq $this->{ENV_SEP} && !$bracelevel) {
	push @things, $current  if length($current);
	$current = '';
    } else {
      $current .= $item;
      if ($item eq '$') { # '
	  $item = shift @text;
	  $current .= $item;
	  if ($item eq '{') {
	    while (1) {
	      $item = shift @text;
	      $current .= $item;
	      last if $item eq '}';
	    }
	  }
	} elsif ($item eq '{') {
	  $bracelevel++;
	} elsif ($item eq '}') {
	  $bracelevel--;
	}
    }
  }
  push @things, $current if length($current);
  return @things;
}

sub filename_components {
  my ($this,$path) = @_;
  return undef unless defined $path;
  my @things=();
  my $current = '';
  my $bracelevel = 0;
  my @text;
  if (length($this->{DIR_SEPA}))  {
    @text = split(/(\$|\{|\}|$this->{DIR_SEP}|$this->{DIR_SEPA})/,$path);
  } else {
    @text = split(/(\$|\{|\}|$this->{DIR_SEP})/,$path);
  }
  while (@text) {
    my $item = shift @text;
    if ($this->IS_DIR_SEP($item) && !$bracelevel) {
	push @things, $current  if length($current);
	$current = '';
    } else {
      $current .= $item;
      if ($item eq '$') { # '
	  $item = shift @text;
	  $current .= $item;
	  if ($item eq '{') {
	    while (1) {
	      $item = shift @text;
	      $current .= $item;
	      last if $item eq '}';
	    }
	  }
	} elsif ($item eq '{') {
	  $bracelevel++;
	} elsif ($item eq '}') {
	  $bracelevel--;
	}
    }
  }
  push @things, $current if length($current);
  return @things;
}

#tilde.c

sub tilde_expand {
  my ($this,$name) = @_;
  return $name unless $name =~ /^~/;
  if ($name eq '~') {
    return $ENV{'HOME'} || '.';
  } elsif (length($name)>1 && $this->IS_DIR_SEP(substr($name,1,1))) {
    my $home = $ENV{'HOME'} || '.';
    if ($this->IS_DIR_SEP(substr($home,0,1)) &&  $this->IS_DIR_SEP(substr($home,1,1))) {
      $home =~ s/^.//;
    }
    if ($this->IS_DIR_SEP(substr($home,scalar($home)-1,1))) {
      $home =~ s/.$//;
    }
    return $home . substr($name,1);
  } elsif (length($name)>1) {
    my $c = 1;
    my $nextchar = substr($name,$c,1);
    my $user = '';
    while (length($name)>$c && !($this->IS_DIR_SEP($nextchar))) {
      $user .= $nextchar;
      $c++;
      $nextchar = substr($name,$c,1);
    }
    my ($xname,$passwd,$uid,$gid,
	$quota,$comment,$gcos,$home) = getpwnam($user);
    $home = '.' unless defined $home;
    if ($this->IS_DIR_SEP(substr($home,0,1)) &&  $this->IS_DIR_SEP(substr($home,1,1))) {
      $home =~ s/^.//;
    }
    if ($this->IS_DIR_SEP(substr($home,scalar($home)-1,1))) {
      $home =~ s/.$//;
    }
    return  $home . substr($name,$c);
  } else {
    return $name;
  }
}

#elt-dirs.c

sub dir_list_add {
  my ($this,$l,$dir) = @_;
  my $last_char = substr($dir,length($dir)-1,1);
  my $saved_dir = ($this->IS_DIR_SEP($last_char) || $this->IS_DEVICE_SEP($last_char)
		   ? $dir : $dir . $this->{DIR_SEP});
  if (ref($l) eq 'ARRAY') {
    $l = [@$l,$saved_dir];
  } elsif ($l) {
    $l = [$l,$saved_dir];
  } else {
    $l = [$saved_dir];
  }
  return $l;
}

sub checked_dir_list_add {
  my ($this,$l,$dir) = @_;
  if(-d $dir) {
    return $this->dir_list_add($l,$dir);
  } else {
    return $l;
  }
}

sub cache {
  my ($this,$key,$value) = @_;
  if ($this->{MONOCASE_FILENAMES}) { $key = lc($key); }
  $this->{the_cache}->{$key} = $value;
}

sub cached {
  my ($this,$key);
  if ($this->{MONOCASE_FILENAMES}) { $key = lc($key); }
  return $this->{the_cache}->{$key} if defined $this->{the_cache}->{$key};
  return undef;
}


sub do_subdir {
  my ($this,$str_list_ptr,$elt,$elt_length,$post) = @_;
  my $dir; 
  my $dirent;
  my $nl = length($elt);
  my $name = substr($elt,0,$nl+$elt_length);
  my $dirx = opendir ($dir,$name);
  return unless defined $dirx && $dirx;
  if ($post eq '') {
    $str_list_ptr = $this->dir_list_add ($str_list_ptr, $name);
  }  else { 
    $str_list_ptr = $this->expand_elt ($str_list_ptr, $name . $post, $elt_length);
  }
  while (my $e = readdir ($dir)) {
    if ($e !~ /^\./) {
      my $links = $this->dir_links ($name.$e);
      if ($links >= 0) {
	my $potential_len = length($name.$e);
	if ($post ne '') {
	  $str_list_ptr = $this->expand_elt ($str_list_ptr, $name.$e.$this->{DIR_SEP}.$post, $potential_len);
	}
	if ($links > 2) {
	  $str_list_ptr = $this->do_subdir ($str_list_ptr, $name.$e.$this->{DIR_SEP}, $potential_len, $post);
	} elsif ($post eq '') {
	  $str_list_ptr = $this->dir_list_add ($str_list_ptr, $name.$e.$this->{DIR_SEP});
	}
      }
    }
  }
  closedir ($dir);
  return $str_list_ptr;
}

sub expand_elt {
  my ($this,$ret,$elt,$start) = @_;
  my @elt = split(//,$elt);
  my $dir = $start;
  my $post;
  while ($dir<length($elt)-1) {
    if ($this->IS_DIR_SEP($elt[$dir])) {
      if ($this->IS_DIR_SEP($elt[$dir+1])) {
	for ($post = $dir + 1; (defined $elt[$post]) && $this->IS_DIR_SEP ($elt[$post]); $post++) {}; 
	my $postfixed = substr($elt,$post) || ''; 
	my $val = $dir - length($elt) + 1;
	$ret = $this->do_subdir ($ret, $elt, $val, $postfixed); 
	return $ret;

      }
    }
    $dir ++;
  }
  $ret = $this->checked_dir_list_add ($ret,$elt);
  return $ret;
}

sub normalize_path {
  my ($this,$elt) = @_;
  my $ret=1;
  my $chartest = "$this->{DIR_SEP}$this->{DIR_SEPA}";
  if ($this->NAME_BEGINS_WITH_DEVICE($elt)) {
    $elt =~ s/^([A-Z])/lc($1)/e;
    $elt =~ s/^(..)([$chartest])[$chartest]*/$1$2/;
    $ret = 2;
  } elsif ($this->IS_UNC_NAME($elt)) {
    $elt =~ s/^([$chartest])\1([^$chartest]+)\1+/$1$1$2$1/;
    $ret = length($1);
  } else {
    $elt =~ s/^([$chartest])[$chartest]*/$1/;
  }
  if ($this->DEBUG_P ($DEBUG_STAT)) {
    DEBUG ("kpse_normalize_path ($elt) => $ret\n");
  }
  if (wantarray()) {
    return ($elt,$ret);
  } else {
    return $elt;
  }
}

sub element_dirs {
  my ($this,$elt) = @_;
  my $ret = [];
  $ret = $this->cached($elt);
  return $ret if defined $ret;
  $ret = $this->expand_elt($ret,($this->normalize_path($elt)));  
  if ($this->DEBUG_P ($DEBUG_EXPAND)) {
    DEBUG ("path element $elt =>");
    if ($ret) {
      foreach my $e (@$ret) {
	DEB (" $e");
      }
    }
    DEB("\n");
  }

  $this->cache($elt,$ret);
  return $ret;
}





#variable.c


sub var_value {
  my ($this,$var) = @_;
  my $ret = $ENV{"$var." . $this->{program_name}};
  $ret = $ENV{"${var}_" . $this->{program_name}} if (!defined $ret) && 
	defined $ENV{"${var}_" . $this->{program_name}};
  $ret = $ENV{"$var"} if (!defined $ret) && defined $ENV{"$var"};
  $ret = $this->cnf_get($var)  unless defined $ret;
  $ret = $this->var_expand($ret) if defined $ret && $ret;
  if ($this->DEBUG_P ($DEBUG_VARS)) {
    DEBUGF("variable: $var = %s\n", ($ret ? $ret : "(nil)"));
  }
  return $ret;
}


sub _expand {
  my ($this,$var) = @_;
  my $ret = undef;
  if ($this->{expansions}->{$var}) {
    # self-ref
	return '';
  } else {
    $ret = $ENV{"${var}_" . $this->{program_name}} 
      if defined $ENV{"${var}_" . $this->{program_name}};
    $ret = $ENV{"$var"} if (!defined $ret) && defined $ENV{"$var"};
    $ret = $this->cnf_get($var)  unless defined $ret;
    if (defined $ret && $ret) {
      $this->{expansions}->{$var} =1;
      $ret = $this->var_expand($ret);
      $this->{expansions}->{$var} =0;
    }
    return $ret if defined $ret;
    return '';
  }
}


sub var_expand {
  my ($this,$src) = @_;
  my $ret; 
  my $expansion='';
  my @src= split(//,$src);
  while (@src) {
    my $s = shift @src;
    if ($s eq '$') { # '
      return $expansion . $s unless @src;
      $s = shift @src;
      if ($s =~ /[a-zA-Z0-9_]/) {
	my $var = '';
	while($s =~ /[a-zA-Z0-9_]/) {
	  $var .= $s;
	  $s = shift @src || '';
	};
	$expansion .= $this->_expand($var) if length($var);
	unshift @src,$s if length($s);
      } elsif ($s eq '{') {
	$s = shift @src;
	my $var = '';
	while ($s ne '}') {
	  $var .= $s;
	  $s = shift @src || '';
	};
	$expansion .= $this->_expand($var) if length($var);
      } else {
	# warning: : Unrecognized variable construct `$%c'", src, *s);
      };
    } else {
      $expansion .= $s;
    }
  }
  return $expansion;
}

#expand.c:

sub expand {
  my ($this,$s) = @_;
  return $this->tilde_expand($this->var_expand($s));
}

sub expand_dot {
  my ($this,$path) = @_;
  my $dot = $ENV{KPSE_DOT} || '';
  my $env = $this->{ENV_SEP};
  return $path if (!$dot);
  my $ret = '';
  $^O =~ /msdos/i && $dot =~ s/^\/\/([A-Za-z])\//$1:\//;
  foreach my $elt ($this->path_elements($path)) {
    if ($this->absolute_p($elt,0) || $elt =~ /^!!/) {
      $ret .= $elt . $env;
    } elsif ($elt eq '.') {
      $ret .= $dot . $env;
    } else {
      next if $^O =~ /vms/i;
      if ($elt =~  /^\.(.)/ && $this->IS_DIR_SEP($1)) {
	$ret .= $dot . substr($elt,1) . $env;
      } else {
	$ret .= $dot . $this->{DIR_SEP} . $elt . $env;
      }
    } 
  }
  $ret =~ s/$env$//;
  return $ret;
}

sub brace_expand_element {
  my ($this,$elt) = @_;
  my $ret = '';
  foreach my $i ($this->_brace_expand($elt)) {
    my $x = $this->expand($i);
    if ($x ne $i) {
      $x = $this->brace_expand_element($x);
    }
    $ret .= $x . $this->{ENV_SEP};
  }
  $ret =~ s/$this->{ENV_SEP}$//;
  return $ret;
}


sub brace_expand {
  my ($this,$path) = @_;
  my $xpath = $this->var_expand($path);
  my $ret = '';
  foreach my $elt ($this->path_elements($xpath)) {
    my $exp = $this->brace_expand_element($elt);
    $ret .= $exp . $this->{ENV_SEP};
  }
  $ret =~ s/$this->{ENV_SEP}$//;
  return $this->expand_dot($ret);
}

sub path_expand {
  my ($this,$path) = @_;
  my $xpath = $this->brace_expand($path);
  my $ret ;
  foreach my $elt ($this->path_elements($xpath)) {
    $elt =~ s/^!!//;
    my $dirs =$this->element_dirs($elt) || [];
    foreach my $dir (@$dirs) {
      my $dirlen = length($dir);
      if ($dirlen==1 || ($dirlen==3 && 
			 $this->NAME_BEGINS_WITH_DEVICE($dir) &&
                         $this->IS_DIR_SEP(substr($dir,2,1)))) {
	$ret .= $dir . $this->{ENV_SEP};
      } else {
	$ret .= $dir;
	chop $ret; # remove traling /
	$ret .= $this->{ENV_SEP};
      }
    }
  }
  $ret =~ s/$this->{ENV_SEP}$//;
  return $ret;
}

sub i_expand {
  my ($this,$text) = @_;
  my @results = ();
  # ${...} signifies a variable, so only match "{" if not
  # preceded by a dollar sign.
  if ($text =~ /(?<!\$)\{.*\}/) {
    $text =~ /^(.*?)(?<!\$)\{([^\{\}]+)\}(.*)$/;
    my $pre = $1; my $post = $3;  my $main = $2;
    my @parts = split(/[$this->{ENV_SEP},]/,$main,-1); # we need empty sections as well
    foreach my $part (@parts) {
      # the array expansion here is not efficient (but short code)
      @results = (@results,$this->i_expand("$pre$part$post"));
    }
    return @results;
  } else {
    return ($text);
  }
}

sub _brace_expand {
  my ($this,$text) = @_;
  my %seen = ();
  return grep {  $seen{$_} += 1; $seen{$_}< 2;} $this->i_expand($text);
}


#proginit.c

sub init_prog {
  my ($this,$prefix,$dpi,$mode,$fallback) = @_;
  if (defined $ENV{$prefix . "MAKEPK"}) {
    $this->set_program_enabled ($this->{pk_format}, 1, $this->{src_env});
    $this->set_program_enabled ($this->{any_glyph_format}, 1, $this->{src_env});
    $this->{format_info}->[$this->{pk_format}]->{program}
      = $this->{format_info}->[$this->{any_glyph_format}]->{program} = $ENV{$prefix . "MAKEPK"};
  }
  $this->{format_info}->[$this->{pk_format}]->{override_path}
    = $this->{format_info}->[$this->{gf_format}]->{override_path}
      = $this->{format_info}->[$this->{any_glyph_format}]->{override_path}
	= $this->{format_info}->[$this->{tfm_format}]->{override_path} = $ENV{ $prefix . "FONTS"};
  $this->{format_info}->[$this->{tex_ps_header_format}]->{override_path} = $ENV{ $prefix . "HEADERS"};
  $this->init_fallback_resolutions($prefix . "SIZES");
  $this->xputenv('MAKETEX_BASE_DPI',$dpi);
  $this->{fallback_font} = $fallback;
  $this->xputenv ("MAKETEX_MODE", $mode ? $mode : $this->{DIR_SEP});
}


#xdirname.c

sub xdirname {
  my ($this,$name) = @_;
  return unless defined $name && scalar($name);
  my $chartest = "$this->{DIR_SEP}$this->{DIR_SEPA}";
  if ($this->NAME_BEGINS_WITH_DEVICE($name)) {
    $name =~ s/^(.*?){2,}[^$chartest]*[$chartest]?$/$1/;
    $name .= "." if length($name) == 2;
    return $name;
  } elsif ($this->IS_UNC_NAME($name)) {
    $name =~ s/^([$chartest]\1[^$chartest]+\1.*?)[^$chartest]*[$chartest]?$/$1/;
    return $name;
  } else {
    $name =~ s/^(.*?[$chartest]?)[^$chartest]*[$chartest]?$/$1/;
    return $name;
  }
}


sub xbasename {
  my ($this,$name) = @_;
  my $chartest = "$this->{DIR_SEP}$this->{DIR_SEPA}";
  $name =~ /([^$chartest]*)$/;
  return $1;
}


#progname.c

sub set_program_name {
  my ($this,$argvnull,$progname) = @_;
  #TODO: no STDERR Redirection is done
  my $s = $ENV{KPATHSEA_DEBUG};
  $this->{kpathsea_debug} = $s if defined $s && $s;
  $this->{program_invocation_name} = $FindBin::RealScript;
  $this->xputenv ("SELFAUTOLOC", $FindBin::Bin);
  my $sdir_parent = $this->xdirname ($FindBin::Bin);
  $this->xputenv ("SELFAUTODIR", $sdir_parent);
  my $sdir_grandparent = $this->xdirname ($sdir_parent);
  $this->xputenv ("SELFAUTOPARENT", $sdir_grandparent);
  $this->{program_invocation_short_name} = $this->xbasename($this->{program_invocation_name});
  if (defined $progname && $progname) {
    $this->{program_name} = $progname;
  } else {
    $this->{program_name} = $this->{program_invocation_short_name};
    $this->{program_name} =~ s/(\.exe|\.pl)$//;
  }
}

sub dir_list_search {
  my ($this,$dirs,$name,$searchall) = @_;
  my $ret = [];
  foreach my $dir (@$dirs) {
    if ($this->readable_file ($dir.$name)) {
      push @$ret, $dir . $name;
      return $ret unless $searchall;
    }
  }
  return $ret;
}

sub absolute_search {
  my ($this,$name) = @_;
  if ($this->readable_file ($name)) {
    return [$name];
  } else {
    return [];
  }
}

sub _path_search {
  my ($this,$path,$name,$must_exist,$all) = @_;
  my $found=0;
  my $ret_list=[];
  foreach my $elt ($this->path_elements($path)) {
    my $allow_disk_search = 1;
    $allow_disk_search = 0 if $elt =~ s/^!!//;
    $elt = $this->normalize_path($elt);
    $found = ($this->{not_first_search} ? $this->db_search ($name, $elt, $all) : undef);
    if ($allow_disk_search
	&& ($must_exist && !(defined $found))) {
      my $dirs = $this->element_dirs ($elt);
      if (defined $dirs && $dirs) {
        if (!defined $found) {
	  print "search disk\n";
          $found = $this->dir_list_search ($dirs, $name, $all);
	}
      }
    }
    if (defined $found && $found) {
      if ($all) {
	$ret_list = [@$ret_list,@{$found}];
      } else {
	$ret_list = [@$ret_list,$found->[0]];
	last;
      }
    }
  }
  return $ret_list;
}

sub search {
  my ($this,$path,$original_name,$must_exist,$all) = @_;
  my $name = $this->expand ($original_name);
  my $absolute_p = $this->absolute_p ($name,1);
  if ($this->DEBUG_P ($DEBUG_SEARCH)) {
    DEBUG ("start search(file=$name, must_exist=$must_exist, find_all=$all, path=$path).\n");
  }
  my $ret_list = ($absolute_p ? $this->absolute_search ($name)
                        : $this->_path_search ($path, $name, $must_exist, $all));
#  print Dumper($ret_list);
  unless ($this->{not_first_search}) {
    $this->{not_first_search} = 1;
  } else {
    if ($this->DEBUG_P ($DEBUG_SEARCH)) {
      DEBUG ("search($original_name) =>");
    }
    $this->log_search ($ret_list);
    if ($this->DEBUG_P ($DEBUG_SEARCH)) {
      DEB("\n");
    }
  }
  return $ret_list;
}


sub log_search {
  my ($this,$filenames) = @_;
  if (!(defined $this->{not_first_time}) &&
      !(defined $this->{log_file})) {
    my $log_name = $this->var_value("TEXMFLOG");
    if ($log_name) {
      $this->{log_file} = $this->xfopen ($log_name, ">");
      if (!$this->{log_file}) {
        perror ($log_name);
	exit 1;
      }
    }
    $this->{not_first_time} = 1;
  }
  if ($this->DEBUG_P ($DEBUG_SEARCH) || defined $this->{log_file}) {
    foreach my $filename (@$filenames) {
      if($this->{log_file} && $this->absolute_p ($filename, 0)) {
	print ($this->{log_file},time() . " $filename\n");
      }
      if ($this->DEBUG_P ($DEBUG_SEARCH)) {
	DEB(" $filename") if $filename;
      }
    }
  }
}


sub path_search {
  my ($this,$path,$name,$must_exist) = @_;
  my $ret_list = $this->search($path,$name,$must_exist,0)->[0];
  return $ret_list;
}

sub all_path_search {
  my ($this,$path,$name,$must_exist) = @_;
  $must_exist = 1 unless defined $must_exist;
  my $ret_list = $this->search($path,$name,$must_exist,1);
  return $ret_list;
}

#db.c:

sub ignore_dir_p {
  my ($this,$dirname) = @_;
  while ($dirname =~ s/(.)\.(.)//) {
    return 1 if $this->IS_DIR_SEP($1) && !$this->IS_DIR_SEP($2);
  }
  return 0;
}

sub db_build {
  my ($this,$table,$db_filename) = @_;
  my $line;
  my ($dir_count, $file_count, $ignore_dir_count) = (0,0,0);
  my $len = length ($db_filename) - length ($this->{DB_NAME}) ;
  my $top_dir = substr($db_filename,0,$len);
  my $cur_dir;
  my $db_file = $this->xfopen($db_filename,"<");
  if ($db_file) {
    while ($line = $this->read_line ($db_file)) {
      chomp $line;
      $len = length ($line);
      if ($len > 0 && $line =~ /:$/ && $this->absolute_p ($line,1)) {
        if (!$this->ignore_dir_p ($line)) {
          $line =~ s/:$/$this->{DIR_SEP}/;
          $cur_dir = ($line =~ s/^\..// ? $top_dir  . $line : $line);
          $dir_count++;
        } else {
          $cur_dir = undef;
          $ignore_dir_count++;
        }
      } elsif ($line && $cur_dir && !($line eq '.' || $line eq '..')) {
        $this->hash_insert_normalized($table,$line,$cur_dir);
        $file_count++;
      }
    }
    $this->xfclose ($db_file, $db_filename);
    if ($file_count == 0) {
#      WARNING ("kpathsea: No usable entries in $db_filename");
#      WARNING ("kpathsea: See the manual for how to generate ls-R");
      $db_file = undef;
    } else {
      push @{$this->{db_dir_list}}, $top_dir;
    }
    if ($this->DEBUG_P ($DEBUG_HASH)) {
      DEBUG("$db_filename: $file_count entries in $dir_count directories ($ignore_dir_count hidden).\n");
      DEBUG("ls-R hash table:");
      $this->hash_print ($table, 1);
    }
    return 1;
  }
  return 0;
}

sub db_insert {
  my ($this,$passed_fname) = @_;
    if ($this->{db}) {
      my $baseptr = $this->xbasename ($passed_fname);
      my $file_part = $baseptr;
      my $dir_part = $passed_fname; # /* That leaves the dir, with the trailing /.  */
      $dir_part =~ s/$file_part$//;
      $this->hash_insert ('db', $file_part, $dir_part);
  }
}

# Return true if FILENAME could be in PATH_ELT, i.e., if the directory
#   part of FILENAME matches PATH_ELT.  Have to consider // wildcards

sub match {
  my ($this,$filename,$path_elt) = @_;
  return 1 if ($path_elt =~ /$this->{DIR_SEP}$this->{DIR_SEP}$/);
  if (length($this->{DIR_SEPA})) {
    return 1 if ($path_elt =~ /$this->{DIR_SEPA}$this->{DIR_SEPA}$/);
  }
  my $dirsec = $this->xdirname($filename);
  return 1 unless $dirsec;
  # this is actually wrong if we have more than 1 part:
  # *Any* '//' in $path_elt
  # should return true, since *any* subdir could be present
  # at that point, including the user-specified one.
  foreach my $part (split(/$this->{DIR_SEP}$this->{DIR_SEP}/,$path_elt)) {
    if ($this->{MONOCASE_FILENAMES}) {
      return 1 if $part =~ /$dirsec$/i;
    } else {
      return 1 if $part =~ /$dirsec$/;
    }
  }
  if (length($this->{DIR_SEPA})) {
    foreach my $part (split(/$this->{DIR_SEPA}$this->{DIR_SEPA}/,$path_elt)) {
      if ($this->{MONOCASE_FILENAMES}) {
	return 1 if $part =~ /$dirsec$/i;
      } else {
	return 1 if $part =~ /$dirsec$/;
      }
    }
  }
  return 0;
}

sub elt_in_db {
  my ($this,$db_dir,$path_elt) = @_;
  if ($this->{MONOCASE_FILENAMES}) {
    return 1 if $path_elt =~ /^$db_dir/i;
  } else {
    return 1 if $path_elt =~ /^$db_dir/;
  }
  return 0;
}


sub alias_build {
  my ($this,$table,$alias_filename) = @_;
  my ($line,$real,$alias);
  my $count = 0;
  my $alias_file=$this->xfopen($alias_filename,"<");
  if ($alias_file) {
    while ($line = $this->read_line($alias_file)) {
      next unless $line;
      next if $line =~ /^[\#%]/;
      my ($real,$alias) = split(/ /,$line);
      if (defined $real && defined $alias && length($real) && length($alias)) {
	$this->hash_insert_normalized($table,$alias,$real);
	$count++;
      }
    }
    if ($this->DEBUG_P ($DEBUG_HASH)) {
      DEBUG("$alias_filename: $count aliases.\n");
      DEBUG("alias hash table:");
      $this->hash_print ($table, 1);
    }
    $this->xfclose($alias_file,$alias_filename);
    return 1;
  }
  return 0;
}


sub init_db {
  my ($this) = @_;
  my $db_path = $this->init_format ($this->{db_format});
  my $db_files = $this->all_path_search ($db_path, $this->{DB_NAME});
  foreach my $file (@$db_files) {
    $this->db_build ('db', $file);
  }
  $db_files = $this->all_path_search ($db_path, $this->{ALIAS_NAME});
  foreach my $file (@$db_files) {
    $this->alias_build ('alias_db', $file);
  }
}


sub db_search {
  my ($this,$name,$path_elt,$all) = @_;
  return undef unless defined $this->{db};
  if ($name =~ /^(.+?)\/(.+?)$/) {
    $path_elt .= "/$1";
    $name = $2;
  }
  my $relevant = 0;
  foreach my $elt (@{$this->{db_dir_list}}) {
    $relevant = $this->elt_in_db($elt,$path_elt);
    last if $relevant;
  }
  return undef unless $relevant;
  my $aliases = $this->hash_lookup('alias_db',$name);
  push @$aliases, $name;
  my $found = 0;
  my $ret = [];
  foreach my $r (@$aliases) {
    my $orig_dirs = my $db_dirs = $this->hash_lookup('db',$r);
#    print Dumper("dbdirs", $db_dirs);
    foreach my $dbdir (@$db_dirs) {
      my $db_file = $dbdir . $r;
      my $matched = $this->match($db_file,$path_elt);
      if ($this->DEBUG_P ($DEBUG_SEARCH)) {
        DEBUG ("db:match($db_file,$path_elt) = $matched\n");
      }
      if ($matched) {
	if ($this->readable_file($db_file)) {
	  $found = $db_file;
	} else {
	  foreach my $a (@$aliases) {
	    my $atry = $dbdir . $a;
	    if ($this->readable_file($atry)) {
	      $found = $atry;
	      last;
	    }
	  }
	}
	if ($found) {
	  push @$ret,$found;
	  return $ret if !$all;
	}
      }
    }
  }
  return $ret if scalar(@$ret);
  return undef;
}



sub map_file_parse {
  my ($this,$map_filename) = @_;
  my $map_lineno = 0;
  my $f = $this->xfopen($map_filename,"<");
  while (my $l = $this->read_line($f)) {
    $l =~ s/(%|\@c).*$//;
    $map_lineno++;
    next unless $l;
    my ($filename,$alias) = split(/ /,$l);
    if ($filename eq 'include') {
      next unless $alias;
      my $include_fname = $this->path_search ($this->{map_path}, $alias, 0);
      next unless $include_fname;
      $this->map_file_parse($include_fname);
    } else {
      next unless $alias;
      $this->hash_insert_normalized ('map', $alias, $filename);
    }
  }
  $this->xfclose($f,$map_filename);
}

sub read_all_maps {
  my ($this) = @_;
  $this->{map_path} = $this->init_format($this->{fontmap_format});
  my $filenames = $this->all_path_search($this->{map_path},$this->{MAP_NAME});
  foreach my $file (@$filenames) {
    $this->map_file_parse($file);
  }
}

sub fontmap_lookup {
  my ($this,$key) = @_;
  $this->read_all_maps() unless defined $this->{map};
  my $suffix = $this->find_suffix($key);
  my $ret = $this->hash_lookup('map',$key);
  $ret = $this->{map}->{$key} if $this->{map}->{$key};
  if (!defined $ret) {
    if ($suffix) {
      my $base_key = $this->remove_suffix($key);
      $ret = $this->hash_lookup('map',$base_key);
    }
  }
  if ((defined $ret) && $suffix) {
    map {$_ = $this->extend_filename($_,$suffix)} @$ret;
  }
  return $ret;
}

#kdefault.c

sub expand_default {
  my ($this,$path,$fallback) = @_;
  my $expansion = $fallback unless defined $path && length($path);
  if ($this->IS_ENV_SEP(substr($path,0,1))) {
    $expansion = (length($path)>1 ? $fallback . $path : $fallback);
  } elsif ($this->IS_ENV_SEP(substr($path,length($path)-1,1))) {
    $expansion = $path . $fallback;
  } elsif ($path =~ /^(.*?)$this->{ENV_SEP}$this->{ENV_SEP}(.*)$/) {
    $expansion = $1. $this->{ENV_SEP} . $fallback . $this->{ENV_SEP} . $2;
  } else {
    $expansion = $path;
  }
  return $expansion;
}

#tex-file.c

sub init_fallback_resolutions {
  my ($this,$envvar) = @_;
  my $size_str = $ENV{($ENV{$envvar} ? $envvar : "TEXSIZES")} || '';
  my $size_count = 0;
  my $default_sizes = ($this->{fallback_resolutions_string} 
		       ? $this->{fallback_resolutions_string}
		       : $this->{DEFAULT_FONT_SIZES});
  my $size_list = $this->expand_default($size_str,$default_sizes);
  my $prevsize=0;
  my $ret = [];
  foreach my $size ($this->path_elements($size_list)) {
    next unless $size;
    next if $size<$prevsize;
    push @$ret,$size;
    $prevsize = $size;
  }
  $this->{fallback_resolutions} = $ret;
}

sub set_program_enabled {
  my ($this,$fmt,$value,$level) = @_;
  if ($level > $this->{format_info}->[$fmt]->{program_enable_level}) {
    $this->{format_info}->[$fmt]->{program_enabled_p} = $value;
    $this->{format_info}->[$fmt]->{program_enable_level} = $level;
  }
}

sub maketex_option {
  my ($this,$fmtname,$value) = @_;
  my $fmt  = $this->{last_format};
  my $testname = $fmtname;
  $testname = lc ($testname) if $this->{MONOCASE_FILENAMES};
  if      ($testname eq 'pk') {
    $fmt = $this->{pk_format};
  } elsif ($testname eq 'mf') {
    $fmt = $this->{mf_format};
  } elsif ($testname eq 'tex') {
    $fmt = $this->{tex_format};
  } elsif ($testname eq 'tfm') {
    $fmt = $this->{tfm_format};
  } elsif ($testname eq 'ofm') {
    $fmt = $this->{ofm_format};
  } elsif ($testname eq 'ocp') {
    $fmt = $this->{ocp_format};
  }
  if ($fmt != $this->{last_format}) {
    $this->set_program_enabled($fmt,$value,$this->{src_cmdline});
  }
}

sub EXPAND_DEFAULT {
  my ($this,$info,$try_path,$source_string) = @_;
  if ($try_path) {
    $info->{raw_path} = $try_path;
    my $a = $this->expand_default($try_path,$info->{path});
    $info->{path} = $a;
    $info->{path_source} = $source_string;
  }
}


sub init_path {
  my ($this,$infono,$default_path,@ap) = @_;
  my $info = $this->{format_info}->[$infono];
  my $var = undef;
  my $env_name = '';
  $info->{default_path} = $default_path;
  while ($env_name = shift @ap) {
    if (!defined $var) {
      my $evar = $env_name . "." . $this->{program_name};
      my $env_value = $ENV{$evar};
      if (defined $env_value && $env_value) {
	$var = $evar;
      } else {
	$evar = $env_name . "_" . $this->{program_name};
	$env_value = $ENV{$evar};
	if (defined $env_value && $env_value) {
	  $var = $evar;
	} else {
	  $env_value = $ENV{$env_name};
	  if (defined $env_value && $env_value) {
	    $var = $env_name;
	  }
	}
      }
    }
    if (!$info->{cnf_path} && $infono != $this->{cnf_format}) {
      $info->{cnf_path} = $this->cnf_get($env_name);
    }
    last if $var && $info->{cnf_path};
  }
  $info->{path} = $info->{raw_path} = $info->{default_path};
  $info->{path_source} = "compile-time paths.h";
  EXPAND_DEFAULT ($this,$info,$info->{cnf_path},"texmf.cnf");
  EXPAND_DEFAULT ($this,$info,$info->{client_path},"program config file");
  if ($var) {
    EXPAND_DEFAULT ($this,$info,$ENV{$var},"$var environment variable");
  }
  EXPAND_DEFAULT ($this,$info,$info->{override_path},"application override path");
  $info->{path} = $this->brace_expand($info->{path});
  #
  $this->{format_info}->[$infono] = $info;
}

sub set_suffixes {
  my ($this,$format,$alternate,@ap) = @_;
  my $list;
  if ($alternate) {
    $list = $this->{format_info}->[$format]->{alt_suffix};
  } else {
    $list = $this->{format_info}->[$format]->{suffix};
  }
  my $s;
  while ($s = shift @ap) {
    push @$list, $s;
  }
  if ($alternate) {
    $this->{format_info}->[$format]->{alt_suffix} = $list;
  } else {
    $this->{format_info}->[$format]->{suffix} = $list;
  }
}

sub init_maketex {
  my ($this,$fmt,$dflt_prog,@ap) = @_;
  my $f = $this->{format_info}->[$fmt];
  my $prog = ($f->{program}? $f->{program} : $dflt_prog);
  my $progval = $this->var_value(uc($prog));
  $f->{program} = $prog;
  $f->{argc} = scalar(@ap);
  $f->{argv} = [$dflt_prog,@ap];
  if (defined $progval && $progval) {
    $this->set_program_enabled($fmt,($progval==1),$this->{src_client_cnf})
  }
  $this->{format_info}->[$fmt] = $f;
}


sub remove_dbonly {
  my ($this,$path) = @_;
  $path =~ s/$this->{ENV_SEP}!!/$this->{ENV_SEP}/g;
  $path =~ s/^!!//;
  return $path;
}

sub _init_format {
  my ($this,$format,$text,$default_path,@envs) = @_;
  $this->{format_info}->[$format]->{type} = $text;
  $this->init_path($format,$default_path,@envs);
}

sub init_format {
  my ($this,$format) = @_;
  if ($this->{format_info}->[$format]->{path}) {
    return $this->{format_info}->[$format]->{path};
  }
  if ($format == $this->{gf_format}) {
      $this->_init_format ($format,"gf", $this->{DEFAULTS}->{GFFONTS}, @{$this->{GF_ENVS}});
      $this->set_suffixes($format,0,"gf");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{pk_format}) {
      $this->init_maketex ($format, "mktexpk", @{$this->{MKTEXPK_ARGS}});
      $this->_init_format ($format,"pk", $this->{DEFAULTS}->{PKFONTS}, @{$this->{PK_ENVS}});
      $this->set_suffixes($format,0,"pk");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{any_glyph_format}) {
      $this->init_maketex ($format, "mktexpk", @{$this->{MKTEXPK_ARGS}});
      $this->_init_format ($format,"bitmap font", $this->{DEFAULTS}->{GLYPHFONTS}, @{$this->{GLYPH_ENVS}});
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{tfm_format}) {
      $this->init_maketex ($format, "mktextfm");
      $this->_init_format ($format,"tfm", $this->{DEFAULTS}->{TFMFONTS}, @{$this->{TFM_ENVS}});
      $this->set_suffixes($format,0,".tfm");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{afm_format}) {
      $this->_init_format ($format,"afm", $this->{DEFAULTS}->{AFMFONTS}, @{$this->{AFM_ENVS}});
      $this->set_suffixes($format,0,".afm");
    } elsif ($format == $this->{base_format}) {
      $this->init_maketex ($format, "mktexfmt");
      $this->_init_format ($format,"base", $this->{DEFAULTS}->{MFBASES}, @{$this->{BASE_ENVS}});
      $this->set_suffixes($format,0,".base");
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{bib_format}) {
      $this->_init_format ($format,"bib", $this->{DEFAULTS}->{BIBINPUTS}, @{$this->{BIB_ENVS}});
      $this->set_suffixes($format,0,".bib");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
    } elsif ($format == $this->{bst_format}) {
      $this->_init_format ($format,"bst", $this->{DEFAULTS}->{BSTINPUTS}, @{$this->{BST_ENVS}});
      $this->set_suffixes($format,0,".bst");
    } elsif ($format == $this->{cnf_format}) {
      $this->_init_format ($format,"cnf", $this->{DEFAULTS}->{TEXMFCNF}, @{$this->{CNF_ENVS}});
      $this->set_suffixes($format,0,".cnf");
    } elsif ($format == $this->{db_format}) {
      $this->_init_format ($format,"ls-R", $this->{DEFAULTS}->{TEXMFDBS}, @{$this->{DB_ENVS}});
      $this->set_suffixes($format,0,"ls-R");
      $this->{format_info}->[$format]->{path} = 
	$this->remove_dbonly ($this->{format_info}->[$format]->{path});
    } elsif ($format == $this->{fmt_format}) {
      $this->init_maketex ($format, "mktexfmt");
      $this->_init_format ($format,"fmt", $this->{DEFAULTS}->{TEXFORMATS}, @{$this->{FMT_ENVS}});
      $this->set_suffixes($format,0,".fmt");
      $this->set_suffixes($format,1,".efmt", ".efm", ".oft");
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{fontmap_format}) {
      $this->_init_format ($format,"map", $this->{DEFAULTS}->{TEXFONTMAPS}, @{$this->{FONTMAP_ENVS}});
      $this->set_suffixes($format,0,".map");
    } elsif ($format == $this->{mem_format}) {
      $this->init_maketex ($format, "mktexfmt");
      $this->_init_format ($format,"mem", $this->{DEFAULTS}->{MPMEMS}, @{$this->{MEM_ENVS}});
      $this->set_suffixes($format,0,".mem");
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{mf_format}) {
      $this->init_maketex ($format, "mktexmf");
      $this->_init_format ($format,"mf", $this->{DEFAULTS}->{MFINPUTS}, @{$this->{MF_ENVS}});
      $this->set_suffixes($format,0,".mf");
    } elsif ($format == $this->{mft_format}) {
      $this->_init_format ($format,"mft", $this->{DEFAULTS}->{MFTINPUTS}, @{$this->{MFT_ENVS}});
      $this->set_suffixes($format,0,".mft");
    } elsif ($format == $this->{mfpool_format}) {
      $this->_init_format ($format,"mfpool", $this->{DEFAULTS}->{MFPOOL}, @{$this->{MFPOOL_ENVS}});
      $this->set_suffixes($format,0,".pool");
    } elsif ($format == $this->{mp_format}) {
      $this->_init_format ($format,"mp", $this->{DEFAULTS}->{MPINPUTS}, @{$this->{MP_ENVS}});
      $this->set_suffixes($format,0,".mp");
    } elsif ($format == $this->{mppool_format}) {
      $this->_init_format ($format,"mppool", $this->{DEFAULTS}->{MPPOOL}, @{$this->{MPPOOL_ENVS}});
      $this->set_suffixes($format,0,".pool");
    } elsif ($format == $this->{mpsupport_format}) {
      $this->_init_format ($format,"MetaPost support", $this->{DEFAULTS}->{MPSUPPORT}, @{$this->{MPSUPPORT_ENVS}});
    } elsif ($format == $this->{ocp_format}) {
      $this->init_maketex ($format, "mkocp");
      $this->_init_format ($format,"ocp", $this->{DEFAULTS}->{OCPINPUTS}, @{$this->{OCP_ENVS}});
      $this->set_suffixes($format,0,".ocp");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{ofm_format}) {
      $this->init_maketex ($format, "mkofm");
      $this->_init_format ($format,"ofm", $this->{DEFAULTS}->{OFMFONTS}, @{$this->{OFM_ENVS}});
      $this->set_suffixes($format,0, ".ofm", ".tfm");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{opl_format}) {
      $this->_init_format ($format,"opl", $this->{DEFAULTS}->{OPLFONTS}, @{$this->{OPL_ENVS}});
      $this->set_suffixes($format,0,".opl");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
    } elsif ($format == $this->{otp_format}) {
      $this->_init_format ($format,"otp", $this->{DEFAULTS}->{OTPINPUTS}, @{$this->{OTP_ENVS}});
      $this->set_suffixes($format,0,".otp");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
    } elsif ($format == $this->{ovf_format}) {
      $this->_init_format ($format,"ovf", $this->{DEFAULTS}->{OVFFONTS}, @{$this->{OVF_ENVS}});
      $this->set_suffixes($format,0,".ovf");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{ovp_format}) {
      $this->_init_format ($format,"ovp", $this->{DEFAULTS}->{OVPFONTS}, @{$this->{OVP_ENVS}});
      $this->set_suffixes($format,0,".ovp");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
    } elsif ($format == $this->{pict_format}) {
      $this->_init_format ($format,"graphic/figure", $this->{DEFAULTS}->{TEXINPUTS}, @{$this->{PICT_ENVS}});
      $this->set_suffixes($format,1, ".eps", ".epsi");
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{tex_format}) {
      $this->init_maketex ($format, "mktextex");
      $this->_init_format ($format,"tex", $this->{DEFAULTS}->{TEXINPUTS}, @{$this->{TEX_ENVS}});
      $this->set_suffixes($format,0,".tex");
    } elsif ($format == $this->{tex_ps_header_format}) {
      $this->_init_format ($format,"PostScript header", $this->{DEFAULTS}->{TEXPSHEADERS},@{$this->{TEX_PS_HEADER_ENVS}});
      $this->set_suffixes($format,1, ".pro", ".enc");
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{texdoc_format}) {
      $this->_init_format ($format,"TeX system documentation", $this->{DEFAULTS}->{TEXDOCS},@{$this->{TEXDOC_ENVS}});
    } elsif ($format == $this->{texpool_format}) {
      $this->_init_format ($format,"texpool", $this->{DEFAULTS}->{TEXPOOL}, @{$this->{TEXPOOL_ENVS}});
      $this->set_suffixes($format,0,".pool");
    } elsif ($format == $this->{texsource_format}) {
      $this->_init_format ($format,"TeX system sources", $this->{DEFAULTS}->{TEXSOURCES},@{$this->{TEXSOURCE_ENVS}});
    } elsif ($format == $this->{troff_font_format}) {
      $this->_init_format ($format,"Troff fonts", $this->{DEFAULTS}->{TRFONTS},@{$this->{TROFF_FONT_ENVS}});
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{type1_format}) {
      $this->_init_format ($format,"type1 fonts", $this->{DEFAULTS}->{T1FONTS},@{$this->{TYPE1_ENVS}});
      $this->set_suffixes($format,0,".pfa", ".pfb");
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{vf_format}) {
      $this->_init_format ($format,"vf", $this->{DEFAULTS}->{VFFONTS}, @{$this->{VF_ENVS}});
      $this->set_suffixes($format,0,".vf");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{dvips_config_format}) {
      $this->_init_format ($format,"dvips config", $this->{DEFAULTS}->{TEXCONFIG}, @{$this->{DVIPS_CONFIG_ENVS}});
    } elsif ($format == $this->{ist_format}) {
      $this->_init_format ($format,"ist", $this->{DEFAULTS}->{INDEXSTYLE}, @{$this->{IST_ENVS}});
      $this->set_suffixes($format,0,".ist");
    } elsif ($format == $this->{truetype_format}) {
      $this->_init_format ($format,"truetype fonts", $this->{DEFAULTS}->{TTFONTS}, @{$this->{TYPE1_ENVS}});
      $this->set_suffixes($format,0, ".ttf", ".ttc");
      $this->{format_info}->[$format]->{suffix_search_only} = 1;
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{type42_format}) {
      $this->_init_format ($format,"type42 fonts", $this->{DEFAULTS}->{T42FONTS}, @{$this->{TYPE42_ENVS}});
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{web2c_format}) {
      $this->_init_format ($format,"web2c files", $this->{DEFAULTS}->{WEB2C}, @{$this->{WEB2C_ENVS}});
    } elsif ($format == $this->{program_text_format}) {
      $this->_init_format ($format,"other text files",
		   ".".$this->{ENV_SEP}."\$TEXMF/" . $this->{program_name} . "//",
                   uc ($this->{program_name}) . "INPUTS");
    } elsif ($format == $this->{program_binary_format}) {
      $this->_init_format ($format,"other binary files",
		   ".".$this->{ENV_SEP}."\$TEXMF/" . $this->{program_name} . "//",
                   uc ($this->{program_name}) . "INPUTS");
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{miscfonts_format}) {
      $this->_init_format ($format,"misc fonts", $this->{DEFAULTS}->{MISCFONTS}, @{$this->{MISCFONTS_ENVS}});
      $this->{format_info}->[$format]->{binmode} = 1;
    } elsif ($format == $this->{web_format}) {
      $this->_init_format ($format,"web", $this->{DEFAULTS}->{WEBINPUTS}, @{$this->{WEB_ENVS}});
      $this->set_suffixes($format,0,".web");
      $this->set_suffixes($format,1,".ch");
    } elsif ($format == $this->{cweb_format}) {
      $this->_init_format ($format,"cweb", $this->{DEFAULTS}->{CWEBINPUTS}, @{$this->{CWEB_ENVS}});
      $this->set_suffixes($format,0, ".w", ".web");
      $this->set_suffixes($format,1,".ch");
    } else {
      # FATAL1 ("init_format) { Unknown format %d", format);
    }
  if ($this->DEBUG_P ($DEBUG_PATHS)) {
    $this->{format_temp} = $format;
    my $fmt_info = $this->{format_info}->[$this->{format_temp}];
    sub MAYBE {
      my ($this,$val) = @_;
      my $fmt_info = $this->{format_info}->[$this->{format_temp}];
      return $fmt_info->{$val} if ($fmt_info->{$val});
      return "(none)";
    }
    DEBUGF ("Search path for %s files (from %s)\n", $fmt_info->{type}, $fmt_info->{path_source});
    DEBUGF ("  = %s\n", $fmt_info->{path});
    DEBUGF ("  before expansion = %s\n", $fmt_info->{raw_path});
    DEBUGF ("  application override path = %s\n",$this->MAYBE ('override_path'));
    DEBUGF ("  application config file path = %s\n",$this->MAYBE ('client_path'));
    DEBUGF ("  texmf.cnf path = %s\n",$this->MAYBE ('cnf_path'));
    DEBUGF ("  compile-time path = %s\n",$this->MAYBE ('default_path'));
    DEBUG ("  default suffixes =");
    if ($fmt_info->{suffix}) {
      foreach my $ext (@{$fmt_info->{suffix}}) {
	DEB(" $ext");
      }
      DEB ("\n");
    } else {
      DEB(" (none)\n");
    }
    DEBUG  ("  other suffixes =");
    if ($fmt_info->{alt_suffix}) {
      foreach my $ext (@{$fmt_info->{alt_suffix}}) {
	DEB(" $ext");
      }
      DEB ("\n");
    } else {
      DEB(" (none)\n");
    }
    DEBUGF ("  search only with suffix = %d\n",($fmt_info->{suffix_search_only}?1:0));
    DEBUGF ("  numeric format value = %d\n", $format);
    DEBUGF ("  runtime generation program = %s\n",$this->MAYBE ('program'));
    DEBUG  ("  runtime generation command =");
    if ($fmt_info->{argv}) {
      foreach my $arg (@{$fmt_info->{argv}}) {
	DEB(" $arg");
      }
      DEB ("\n");
    } else {
      DEB(" (none)\n");
    }
    DEBUGF ("  program enabled = %d\n", ($fmt_info->{program_enabled_p}?1:0));
    DEBUGF ("  program enable level = %d\n", ($fmt_info->{program_enable_level}?1:0));
  }
  return $this->{format_info}->[$format]->{path};
}


sub find_file {
  my ($this,$name,$format,$must_exist,$all) = @_;
  my $ext;
  $all = 0 unless defined $all;
  my $name_len = 0;
  my $name_has_suffix_already = 0;
  my $mapped_name;
  my $mapped_names = [];
  my $use_fontmaps = ($format == $this->{tfm_format}
		      || $format == $this->{gf_format}
		      || $format == $this->{pk_format}
		      || $format == $this->{ofm_format});
  my $ret = undef;
  if (!defined $this->{format_info}->[$format]->{path}) {
    $this->init_format ($format);
  }
  $name_len = length ($name);
  if ($this->{format_info}->[$format]->{suffix}) {
    foreach my $ext (@{$this->{format_info}->[$format]->{suffix}}) {
      my $suffix_len = length ($ext);
      $name_has_suffix_already = (index($name,$ext) == $name_len-$suffix_len);
      last if $name_has_suffix_already;
    }
  }
  if (!$name_has_suffix_already && $this->{format_info}->[$format]->{alt_suffix}) {
    foreach my $ext (@{$this->{format_info}->[$format]->{alt_suffix}}) {
      my $suffix_len = length ($ext);
      $name_has_suffix_already = (index($name,$ext) == $name_len-$suffix_len);
      last if $name_has_suffix_already;
    }
  }
  if (!$name_has_suffix_already && $this->{format_info}->[$format]->{suffix}) {
    foreach my $ext (@{$this->{format_info}->[$format]->{suffix}}) {
      last if defined $ret;
      my $name_with_suffix = $name . $ext;
      if ($all) {
	$ret = $this->all_path_search ($this->{format_info}->[$format]->{path},$name_with_suffix,0);
      } else {
	$ret = $this->path_search ($this->{format_info}->[$format]->{path},$name_with_suffix,0);
      }
      if (!(defined $ret) && $use_fontmaps) {
        $mapped_names = $this->fontmap_lookup ($name_with_suffix);
        foreach my $mapped_name (@$mapped_names) {
	  last if defined $ret;
	  if ($all) {
	    $ret = $this->all_path_search ($this->{format_info}->[$format]->{path},$mapped_name,0);
	  } else {
	    $ret = $this->path_search ($this->{format_info}->[$format]->{path},$mapped_name,0);
	  }
	}
      }
    }
    if (!(defined $ret) && $this->{format_info}->[$format]->{suffix_search_only} && $must_exist) {
      foreach my $ext (@{$this->{format_info}->[$format]->{suffix}}) {
	last if defined $ret;
	my $name_with_suffix = $name . $ext;
	if ($all) {
	  $ret = $this->all_path_search ($this->{format_info}->[$format]->{path}, $name_with_suffix, 1);
	} else {
	  $ret = $this->path_search ($this->{format_info}->[$format]->{path}, $name_with_suffix, 1);
	}
      }
    }
  }
  if (!(defined $ret)
      &&  ($name_has_suffix_already
	   || 
	   ! $this->{format_info}->[$format]->{suffix_search_only})) {
    if ($all) {
      $ret = $this->all_path_search ($this->{format_info}->[$format]->{path},$name,0);
    } else {
      $ret = $this->path_search ($this->{format_info}->[$format]->{path},$name,0);
    }
    if (!(defined $ret) && $use_fontmaps) {
      $mapped_names = $this->fontmap_lookup ($name);
      foreach my $mapped_name (@$mapped_names) {
	last if defined $ret;
	if ($all) {
	  $ret = $this->all_path_search ($this->{format_info}->[$format]->{path}, $mapped_name, 0);
	} else {
	  $ret = $this->path_search ($this->{format_info}->[$format]->{path}, $mapped_name, 0);
	}
      }
    }
    if (!(defined $ret) && $must_exist) {
      if ($all) {
	$ret = $this->all_path_search ($this->{format_info}->[$format]->{path}, $name,1);
      } else {
	$ret = $this->path_search ($this->{format_info}->[$format]->{path}, $name,1);
      }
    }
  }
  if (!(defined $ret) && $must_exist) {
    $ret = $this->make_tex($format,$name);
  }
  # Guard against:
  #   $ret = ['/opt/tex/fonts/cmr10.tfm', '/opt/tex/fonts/cmr10.tfm']:
  # This 'problem' occurs quite often because expansion tends to create doubled
  # path segments. It could be fixed somewhere earlier, probably, but it doesnt
  # matter much. Just means that we run slighlty less efficient if all entries
  # are needed (which doesnt happen often anyway)
  if (ref($ret) eq 'ARRAY') {
    my %seen = ();
    @$ret = grep {  $seen{$_} += 1; $seen{$_}< 2;} @$ret;
  }
  return $ret;
}

sub open_file {
  my ($this,$name,$type) = @_;
  my $fullname = $this->find_file($name,$type,1);
  my $f = $this->xfopen($fullname,"<");
  unless ($f) {
    exit(1);
  }
  binmode ($f) if $this->{format_info}->[$type]->{binmode};
  return $f;
}


sub reset_program_name {
  my ($this,$progname) = @_;
  return if $this->{program_name} eq  $progname;
  $this->{program_name} = $progname;
  for (my $i = 0; $i != $this->{last_format}; ++$i) {
    if ($i == $this->{cnf_format} || $i == $this->{db_format}) {
      next;
    }
    if (defined $this->{format_info}->[$i]->{path}) {
      $this->{format_info}->[$i]->{path} = undef;
    }
    if (defined $this->{format_info}->[$i]->{cnf_path}) {
      $this->{format_info}->[$i]->{cnf_path} = undef;
    }
  }
}


#tex-glyph.c

sub try_format {
  my ($this,$fontname,$dpi,$format) = @_;
  my $bitmap_specs = [$this->{UNIX_BITMAP_SPEC},$this->{DPI_BITMAP_SPEC}];
  my $path = $this->{format_info}->[$format]->{path};
  unless (defined $path && $path) {
    $path = $this->init_format($format);
  }
  my $sfx = $this->{format_info}->[$format]->{suffix};
  if (defined $sfx && defined $sfx->[0]) {
    $this->xputenv('KPATHSEA_FORMAT',$sfx->[0]);
  }
  my $ret = undef;
  foreach my $must_exist (0..1) {
    last if defined $ret;
    foreach my $spec (@$bitmap_specs) {
      last if defined $ret;
      my $name = $this->var_expand($spec);
      $ret = $this->path_search($path,$name,$must_exist);
    }
  }
  return $ret;
}

sub try_size {
  my ($this,$fontname,$dpi,$format,$glyph_file) = @_;
  my $format_found;
  my $ret = undef;
  my $try_gf = ($format == $this->{gf_format} || $format == $this->{any_glyph_format});
  my $try_pk = ($format == $this->{pk_format} || $format == $this->{any_glyph_format});
  $this->xputenv("KPATHSEA_DPI",$dpi);
  $ret = $try_pk ? $this->try_format ($fontname, $dpi, $this->{pk_format}) : undef;
  if (defined $ret) {
    $format_found = $this->{pk_format};
  } else {
    if ($try_gf) {
      $ret = $this->try_format ($fontname, $dpi, $this->{gf_format});
      $format_found = $this->{gf_format};
    }
  }
  if (defined $ret && defined $glyph_file) { 
    $glyph_file->{name} = $fontname;
    $glyph_file->{dpi} = $dpi;
    $glyph_file->{format} = $format_found;
  }
  return ($ret,$glyph_file);
}

sub BITMAP_TOLERANCE {
  my ($dpi) = @_;
  return $dpi/500+1;
}

sub try_resolution {
  my ($this,$fontname,$dpi,$format,$glyph_file) = @_;
  my $ret = undef;
  ($ret,$glyph_file) = $this->try_size ($fontname, $dpi, $format, $glyph_file);
  if (!defined $ret) {
    my $r;
    my $tolerance = BITMAP_TOLERANCE ($dpi);
    my $lower_bound = int (($dpi - $tolerance) < 0 ? 0 : ($dpi - $tolerance));
    my $upper_bound = int ($dpi + $tolerance);
    for ($r = $lower_bound;  $r <= $upper_bound; $r++) {
      last if defined $ret;
      if ($r != $dpi) {
	($ret,$glyph_file) = $this->try_size ($fontname, $r, $format, $glyph_file);
      }
    }
  }
  return ($ret,$glyph_file);
}


sub try_fontmap {
  my ($this,$fontname,$dpi,$format,$glyph_file) = @_;
  my $mapped_names;
  my $ret = undef;
  $mapped_names = $this->fontmap_lookup ($fontname);
  if ($mapped_names) {
    my $mapped_name;
    my $first_name = $mapped_names->[0];
    foreach my $mapped_name (@$mapped_names) {
      last if defined $ret;
      $this->xputenv ("KPATHSEA_NAME", $mapped_name);
      ($ret,$glyph_file) = $this->try_resolution ($mapped_name, $dpi, $format, $glyph_file);
    }
    if (defined $ret) {
      $fontname = $mapped_name;
    } elsif (!$this->fontmap_lookup ($first_name)) {
      $fontname = $first_name;
    }
  }
  return ($ret,$fontname,$glyph_file);
}

sub try_fallback_resolutions {
  my ($this,$fontname,$dpi,$format,$glyph_file) = @_;
  my $s;
  my ($loc, $max_loc, $lower_loc, $upper_loc);
  my ($lower_diff, $upper_diff);
  my $closest_diff = 2147483647;
  my $ret = undef;
  foreach my $s (0..$#{$this->{fallback_resolutions}}) {
    my $this_diff = abs ($$this->{fallback_resolutions}->[$s] - $dpi);
    if ($this_diff < $closest_diff) {
      $closest_diff = $this_diff;
      $loc = $s;
    }
  }
  return $ret if scalar(@$this->{fallback_resolutions}) == 0;
  $max_loc = scalar(@$this->{fallback_resolutions});
  $lower_loc = $loc - 1;
  $upper_loc = $loc + 1;

  for (;;) {
    my $fallback = $this->{fallback_resolutions}->[$loc];
    if ($fallback != $dpi) {
      ($ret,$glyph_file) = $this->try_resolution ($fontname, $fallback, $format, $glyph_file);
      last if defined $ret;
    }
    $lower_diff = $lower_loc > -1 ?
      $dpi - $this->{fallback_resolutions}->[$lower_loc] :
	2147483647;
    $upper_diff = $upper_loc < $max_loc ? 
      $this->{fallback_resolutions}->[$upper_loc] - $dpi : 
       2147483647;
    last if $lower_diff == 2147483647 && $upper_diff == 2147483647;
    if ($lower_diff < $upper_diff) {
      $loc = $lower_loc;
      $lower_loc--;
    } else {
      $loc = $upper_loc;
      $upper_loc++;
    }
  }
  return ($ret,$glyph_file);
}

sub find_glyph {
  my ($this,$fontname,$dpi,$format) = @_;
  my $glyph_file;
  my $ret = undef;
  my $source = $this->{glyph_source_normal};
  $this->xputenv ("KPATHSEA_NAME", $fontname);
  ($ret,$glyph_file) = $this->try_resolution ($fontname, $dpi, $format, $glyph_file);
  if (!defined $ret) {
    $source = $this->{glyph_source_alias};
    ($ret,$fontname,$glyph_file) = $this->try_fontmap ($fontname, $dpi, $format, $glyph_file);
    if (!(defined $ret) && !$this->absolute_p ($fontname, 1)) {
      $source = $this->{glyph_source_maketex};
      $this->xputenv ("KPATHSEA_DPI", $dpi);
      $ret = $this->make_tex ($format, $fontname);
    }
    if ((defined $ret) && defined $glyph_file) {
      $glyph_file->{dpi} = $dpi;
      $glyph_file->{name} = $fontname;
    } else {
      if ($this->{fallback_resolutions}) {
        ($ret,$glyph_file) = $this->try_fallback_resolutions ($fontname, $dpi, $format, $glyph_file);
      }
      if (!(defined $ret) && $this->{fallback_font}) {
	my $name = $this->{fallback_font};
        $source = $this->{glyph_source_fallback};
        $this->xputenv ("KPATHSEA_NAME", $name);
        ($ret,$glyph_file) = $this->try_resolution ($name, $dpi, $format, $glyph_file);
        if (!(defined $ret) && $this->{fallback_resolutions}) {
	  ($ret,$glyph_file) = $this->try_fallback_resolutions ($name, $dpi, $format, $glyph_file);
	}
      }
    }
  }
  if (defined $glyph_file) {
     $glyph_file->{source} = $source;
   }
  return ($ret,$glyph_file);
}


sub bitmap_tolerance {
  my ($this,$dpia,$dpib) = @_;
  my $tolerance = BITMAP_TOLERANCE ($dpib) ;
  my $lower_bound = ($dpib - $tolerance) < 0 ? 0 : $dpib - $tolerance ;
  my $upper_bound = $dpib + $tolerance ;
  return $lower_bound <= $dpia && $dpia <= $upper_bound 
}

#tex-make.c

sub set_maketex_mag {
  my ($this) = @_;
  my $m;
  my $dpi = $ENV{KPATHSEA_DPI};
  my $bdpi = $ENV{KPATHSEA_BASE_DPI};
  $this->magstep_fix($dpi,$bdpi,\$m);
  my $q;
  if ($m == 0) {
    $q =  sprintf("%d+%d/%d",$dpi / $bdpi, $dpi % $bdpi, $bdpi);
  } else {
    my $sign = '';
    if ($m<0) { $m *= -1; $sign ='-'; }
    $q = sprintf("magstep\\(%s%d.%d\\)", $sign, $m / 2, ($m & 1) * 5);
  }
  $this->xputenv("MAKETEX_MAG",$q);
}



sub misstex {
  my ($this,$format,@args) = @_;
  my $s;
  if ($format != $this->{gf_format}
      && $format != $this->{pk_format}
      && $format != $this->{any_glyph_format}
      && $format != $this->{tfm_format}
      && $format != $this->{vf_format}) {   return; }
  if (!$this->{missfont} && !$this->{make_tex_discard_errors}) {
    my $missfont_name = $this->var_value ("MISSFONT_LOG");
    if (!(defined $missfont_name) || $missfont_name == '1') {
      $missfont_name = "missfont.log";
    } elsif ((defined $missfont_name) && !$missfont_name) {
      undef $missfont_name;
    }
    $this->{missfont} = (defined $missfont_name) ? $this->xfopen ($missfont_name, ">") : undef;
    if (!(defined $this->{missfont}) && $this->var_value ("TEXMFOUTPUT")) {
      $missfont_name = $this->var_value ("TEXMFOUTPUT"). $this->{DIR_SEP} . $missfont_name;
      $this->{missfont} = $this->xfopen ($missfont_name,">");
    }
    if (defined $this->{missfont}) {
      print STDERR "kpathsea: Appending font creation commands to $missfont_name.\n";
    }
  }
  if (defined $this->{missfont}) {
    print ($this->{missfont}, "@args\n");
  }
}

sub maketex {
  my ($this,$format,@args) = @_;
  my $ret = undef;
  my $retv = `@args`;
  chomp $retv;
  my $len = length($retv);
  $ret = ($len ==0 ? undef : $this->readable_file($retv));
  if (defined $ret) {
    $this->db_insert($ret);
  } else {
    $this->misstex($format,@args);
  }
  return $ret;
}

sub make_tex {
  my ($this,$format,$base) = @_;
  my $spec;
  my $ret = undef;
  $spec = $this->{format_info}->[$format];
  if (!defined $spec->{type}) { 
    $this->init_format ($format);
    $spec = $this->{format_info}->[$format];
  }
  if ($spec->{program} && $spec->{program_enabled_p}) {
    if ($format == $this->{gf_format}
        || $format == $this->{pk_format}
        || $format == $this->{any_glyph_format}) {
      $this->set_maketex_mag ();
    }
    my @args;
    my $argnum;
    for ($argnum = 0; $argnum < $spec->{argc}; $argnum++) {
      $args[$argnum] = $this->var_expand ($spec->{argv}->[$argnum]);
    }
    $args[$argnum++] = $base;
    $ret = maketex ($format, @args);
  }
  return $ret;
}

#kpsewhich.c

sub find_dpi {
  my ($this,$s) = @_;
  my $suffix = $this->find_suffix($s);
  if ($suffix) {
    $suffix =~ /^([0-9]+)/;
    return $1;
  } else {
    return undef;
  }
}


sub find_format {
  my ($this,$name,$user_format) = @_;
  $user_format = $this->{last_format} unless defined $user_format;
  my $ret;
  if ((defined $user_format) 
      &&
      $user_format != $this->{last_format}) {
    $ret = $user_format;
  } elsif (lc($name) eq "psfonts.map" ||
           lc($name) eq "pdftex.map" ) {
    $ret = $this->{dvips_config_format};
  } else {
    my $f; 
    my $found = 0;
    for ($f = 0; !$found && $f < $this->{last_format}; $f++) {
      if (!$this->{format_info}->[$f]->{type}) {
        $this->init_format ($f);
      }
      if (!(defined $user_format)) {
          my $ftry = $this->{format_info}->[$f]->{type};
          $found = 1 if $name =~ /$ftry$/;
      }
      foreach my $ext (@{$this->{format_info}->[$f]->{suffix}}){
		last if $found;
		$found = 1 if $name =~ /$ext$/;
      }
      foreach my $ext (@{$this->{format_info}->[$f]->{alt_suffix}}){
		last if $found;
		$found = 1 if $name =~ /$ext$/;
      }
    } 
    $ret = $found ? $f - 1 : $this->{last_format};
  }
  return $ret;
}

sub format_info_string {
  my ($this,$f) = @_;
  my $a = $this->{format_info}->[$f]->{type} . ":";
  foreach my $ext (@{$this->{format_info}->[$f]->{suffix}}) {
    $a .= " $ext";
  }
  foreach my $ext (@{$this->{format_info}->[$f]->{alt_suffix}}) {
    $a .= " $ext";
  }
  return $a;
}

sub init_formats {
  my ($this) = @_;
  my $f=0;
  while ($f < $this->{last_format}) {
    $this->init_format($f);
    $f++;
  }
}

sub init_program {
  my ($this,$progn,$progname,$dpi,$mode,$fallback_font) = @_;
  $this->set_program_name($progn,$progname);
  $this->init_prog(uc($this->{program_name}), $dpi, $mode, $fallback_font);
}

1;

__END__

=head1 NAME

Kpathsea - Perl extension for generic path searching, configuration, and TeX-specific file searching.

=head1 SYNOPSIS

  use Kpathsea;

  my $kpse = Kpathsea->new([$init_path_prefix, $init_tmppath_prefix]);

  # set program name
  # and set environment settings

  $kpse->set_program_name($0,[$progname]);
  $kpse->init_prog($env_prefix, [$dpi, $mode, $fallback_font]);

  # or do both at once

  $kpse->init_program($0 [,$progname,$dpi,$mode,$fallback_font]);


  # expand a variable
  print $kpse->var_expand  ($var);

  # expand brace notation in a variable
  print $kpse->brace_expand($var);

  # expand a variable to full path parts
  print $kpse->path_expand ($var);


  # search for a filename in an absolute path
  my $retval = $kpse->path_search ($path, $name, $must_exist);
  my $retlist = $kpse->all_path_search ($path, $name, $must_exist);

  # convert a format specification like 'type1 fonts' or
  # a filename like 'cmr10.tfm' into an internal file format id number
  my $format_id = $kpse->find_format($formatname_or_filename);

  # initialize a specific internal file format
  $kpse->init_format($format_id);

  # or do all of them at once
  $kpse->init_formats();


  # find a file
  $retval_or_list = $kpse->find_file ($name, $format_id, $must_exist, $all);

  # alternative syntax for bitmap font files
  ($retval,$info) = $kpse->find_glyph ($fontname,$dpi, $format_id);


=head1 ABSTRACT

This module is a pure perl port of the Kpathsea[rch] library, which implements
generic path searching, configuration, and TeX-specific file searching.

From the original documentation:

  The library's fundamental purpose is to return a filename from a list of
  directories specified by the user, similar to what shells do when
  looking up program names to execute.

Besides the perl module itself (Kpathsea.pm) a perl version of the "kpsewhich" 
executable is also included. Not included are the mktex{dir,lsr,mf,nam,pk,tfm,upd}
scripts, nor the kpathsdea documentation, nor any other part of a TeX 
distribution. Therefore, you will probably need a texk-based distribution like
teTeX or fptex to make use of this module. 


=head1 QUICKSTART

Most of the subroutines and internal variables have the same names as their C counterparts, with just a few notable exceptions:

=over

=item *

All kpse_ prefixes have been stripped. As a consequence, some of the non_prefixed C function names now start with an underscore

=item *

C-style lists of lists have become array references

=item *

find_glyph() returns a list of two values and therefore no longer needs the glyph_file info struct
 as a parameter to the function call

=item *

Some of the debug output is slightly different from the C version

=back

=head1 DESCRIPTION

=head2 MODULE VARIABLES

=over

=item $Kpathsea::VERSION

The Module version number

=item $Kpathsea::KPATHSEA_VERSION

The C library version number

=back

=head2 INSTANCE VARIABLES

There are a few variables belonging to the object that you may find interesting:

=over

=item $kpse->{last_format}

The final file format id number

=item $kpse->{format_info}

An array of file type format information hashes, indexed by format id's.

=item $kpse->{kpathsea_debug}

Kpathsea's debug variable (initialized from $ENV{KPATHSEA_DEBUG})

=item $kpse->{program_name};

The program used while searching.

=item $kpse->{CNF_NAME}

The filename for the main configuration file(s): texmf.cnf

=item $kpse->{DB_NAME}

The filename for the database cache file(s): ls-R

=item $kpse->{ALIAS_NAME}

The filename for the alias specification file(s): aliases


=back

=head2 SUBROUTINES

=over

=item $kpse = $class->new([$init_path_prefix, $init_tmppath_prefix])

Constructor. The two optional arguments allow you to give a
prefix value for the two 'compile-time paths' instead of 
("/usr/share" and "/var/tmp/texfonts").

=item $kpse->find_mf ($filename)

=item $kpse->find_mft ($filename)

=item $kpse->find_pict ($filename)

=item $kpse->find_tex ($filename)

=item $kpse->find_tfm ($filename)

=item $kpse->find_ofm ($filename)

=item $kpse->find_vf ($filename)

=item $kpse->find_ovf ($filename)

Shortcuts for find_file(), see below.

=item $kpse->find_pk ($font,name,$dpi)

=item $kpse->find_vf ($font_name,$dpi)

Shortcuts for find_glyph(), see below.

=item $filename = $kpse->remove_suffix($filename)

Strips the last suffix from a filename

=item $filename = $kpse->extend_filename($filename,$suffix)

Adds $suffix to $filename if $filename does not have an extension

=item $filename = $kpse->xbasename($path)

Returns the filename part of $path

=item $parentdir = $kpse->xdirname($path)

Returns the directory part of $path, stripping the trailing
entry + separator (if you give a directory, you will receive the directory's
parent)

=item $variable = $kpse->var_value($variable)

Searches through the environment and the confuration settings
to find the value associated with this variable, and returns it's
value

=item $result = $kpse->var_expand($source)

Expands variables in $source.

=item $result = $kpse->brace_expand($source)

Expands brace notation in $source

=item $result = $kpse->path_expand($source)

Expands the paths in $source

=item  $kpse->set_program_name($0,[$progname]);

Sets the program name that is used in searches. The optional
argument allows you to 'lie' about the program's name if you
really want to use another program's settings.

=item  $kpse->reset_program_name($progname);

Resets the program name to the one you specified and re-initializes
the search routines.

=item  $kpse->init_prog($env_prefix, [$dpi, $mode, $fallback_font]);

The $env_prefix is used to initialize a few environment variable
names (like "$env_prefix . HEADERS"), $dpi is the base value, $mode
= the default metafont mode, $fallback_font is the default font
to use if the requested one couldn't be found.

=item $filename = $kpse->path_search($path,$name,$must_exist,0);

Searches the database for the file named $name in the path $path. 
If $must_exist is true, it also searches the filesystem. 

=item $filenamearray = $kpse->all_path_search($path,$name,$must_exist,0);

Searches the database for all the files named $name in the path $path. 
If $must_exist is true, it also searches the filesystem.
Returns an arrayref containing the files that were found.

=item $kpse->maketex_option ($fmtname,$boolean)

Enables or disables the mktexXXX program $fmtname.

=item $kpse->init_format($format_id)

Initializes the $kpse->{format_info} entry for the numeric format
$format_id

=item $kpse->find_file($name,$format_id,$must_exist,$all)

First it attemps to fix the extension of $name (adding one if
present), Then it takes care of alias expansion, then calls either
path_search() or all_path_search() depending on $all. Return value is
the same as one of these two functions.  Perhaps runs a mktexXXX
program if it still cannot find any files after attempting with &
without suffix and aliases.


=item my ($filename,$info) = $kpse->find_glyph($fontname,$dpi,$format_id);

Like find_file(), but specifically for bitmapped font files that are
perhaps automatically generated on-the-fly.  $fontname should not gave
an extension!  $format_id is $kpse->{gf_format} or $kpse->{pk_format}
or $kpse->{any_glyph_format}.  The found font file may not be
precisely the $dpi that was requested, which is why there is the extra
return value.  The extra returned hashref contains four keys: name,
dpi, format, source


=item $kpse->open_file($name,$format_id)

Calls find_file() with $must_exist == 1 and $all == 0, then opens the
found file (takes a perhaps needed binmode() into account).

=back

=head1 BUGS AND KNOWN ISSUES

=over

=item *

mktexXXX functionality depends solely on perl backtick behavior.

=item *

KPATHSEA_DEBUG_OUTPUT and KPATHSEA_DEBUG_APPEND are ignored (which
means you cannot redirect debug output under win32)

=item *

Symlinks and ARGV[0] are assumed to be handled 'correctly' by the perl
interpreter.

=item *

Filenames are in essence not constrained at all, not even under ms-dos

-item *

Large sections of the perl code are really C code rewritten for the perl
tokenizer. Some sections are very slow because of that.

=back

=head1 SEE ALSO

The documentation for the kpathsea C library. 
L<kpsewhich(1)>
L<http://tex.aanhet.net/perl/Kpathsea>

=head1 AUTHOR

Taco Hoekwater, E<lt>taco@elvenkind.comE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2003 by Taco Hoekwater

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
